from __future__ import annotations

"""
The Session User Object
"""

from pydantic import Field, model_validator, PrivateAttr
from lazyops.types.models import ConfigDict, schema_extra
from lazyops.libs import lazyload
from .base import BaseModel
from .common import UserType, ValidationMethod
from .user_roles import UserRole
from .user_session import UserSession
from .user_data import AZUserData
from .claims import UserJWTClaims, APIKeyJWTClaims
from .auth import AuthObject
from .security import Authorization, APIKey
from .errors import (
    AuthZeroException,
    InvalidTokenException, 
    NoTokenException, 
    InvalidAPIKeyException, 
    InsufficientPermissionsException,
    InvalidAuthorizationException,
    InvalidAPIKeyPrefixException,
    ExpiredAPIKeyData,
    InvalidAPIKeySecretException,
)

from ..utils.lazy import logger, ThreadPooler
from ..utils.helpers import parse_scopes, get_hashed_key
from ..utils.decoders import decode_token

from typing import Optional, List, Dict, Any, Union, Callable, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    
    from fastapi import Request
    from fastapi.background import BackgroundTasks
    from fastapi.datastructures import Headers
    from ..types.auth import APIKeyData
    from ..flows.user_session import UserSessionFlow
    from ..flows.user_data import UserDataFlow

if lazyload.TYPE_CHECKING:
    import tldextract
else:
    tldextract = lazyload.LazyLoad('tldextract')



class CurrentUser(BaseModel):
    """
    The Current User Object
    """

    scope: Optional[str] = None # allow setting the scope at the class level
    attributes: Optional[Dict[str, Any]] = Field(default_factory = dict) # Persistent Data through Session, if enabled
    data: Optional[Dict[str, Any]] = Field(default_factory = dict) # Non-Persistent Data
    session: Optional[UserSession] = None # Session Data
    role: Optional[UserRole] = None # User Role
    auth_obj: Optional[AuthObject] = None
    
    # Configuration Options
    required: Optional[bool] = True # If True, will raise an exception if the user does not pass validation
    verbose: Optional[bool] = None # If True, will log the validation errors

    # Passthrough Configurations for the Auth Flow
    session_enabled: Optional[bool] = None
    request_id_enabled: Optional[bool] = None
    domain_source_enabled: Optional[bool] = None
    role_configure_enabled: Optional[bool] = None


    # Previous init args
    # allow_maintenance: Optional[bool] = False
    # silent: Optional[bool] = None
    # is_system_user: Optional[bool] = False

    # Set From Validation
    user_type: Optional[UserType] = None # User Type
    validation_method: Optional[ValidationMethod] = None
    claims: Optional[Union[UserJWTClaims, APIKeyJWTClaims]] = None
    user_data: Optional[AZUserData] = None
    api_key: Optional[str] = None

    from_session: Optional[bool] = None

    # Sesssion Attributes
    _pre_validate_hooks: Optional[List[Callable]] = PrivateAttr(None)
    _post_validate_hooks: Optional[List[Callable]] = PrivateAttr(None)

    def __init__(self, **data):
        """
        Validates the current user
        """
        self.pre_init(**data)
        super().__init__(**data)
        self.post_init(**data)

    def pre_init(self, **kwargs):
        """
        Pre Init
        - This behavior can be overridden
        """
        pass

    def post_init(self, **kwargs):
        """
        Post Init
        - This behavior can be overridden
        """
        pass

    async def precall_hook(self, request: 'Request', **kwargs):
        """
        Runs the precall hook
        """
        # Defer these until they are called to prevent loading the settings prior to full initialization
        self._pre_validate_hooks, self._post_validate_hooks = self.settings.ctx.get_validation_hooks()
        if self.session_enabled is None: self.session_enabled = self.settings.user_session_enabled
        if self.domain_source_enabled is None: self.domain_source_enabled = self.settings.user_domain_source_enabled
        if self.request_id_enabled is None: self.request_id_enabled = self.settings.user_request_id_enabled
        if self.role_configure_enabled is None: self.role_configure_enabled = self.settings.user_role_configure_enabled
        if self.verbose is None: 
            self.verbose = self.required or self.settings.is_development_env

    def validate_permissions(self):
        """
        Validates the Permissions
        """
        if self.scope is not None and self.scope not in self.claims.scopes:
            raise InsufficientPermissionsException( detail = f'Insufficient Permissions for Scope: {self.scope}')
    
    def validate_scopes(self, scopes: List[str]):
        """
        Validates the Scopes
        """
        
        for scope in scopes:
            if scope not in self.claims.scopes:
                # if not self.required: 
                if self.verbose: logger.warning(f'Failed to Validate Scope: {scope} against {self.claims.scopes}. Present Scopes: {scopes}')
                raise InsufficientPermissionsException(detail = f'Insufficient Permissions for Scope: {scope}')
        
    @property
    def session_flow(self) -> 'UserSessionFlow':
        """
        Gets the User Session Flow
        """
        return self.get_flow('user_session', user_id = self.user_id)
    
    @property
    def user_data_flow(self) -> 'UserDataFlow':
        """
        Gets the User Data Flow
        """
        return self.get_flow('user_data', user_id = self.user_id)
    

    @property
    def user_id(self) -> Optional[str]:
        """
        Returns the User ID or API Client ID
        """
        return self.claims.sub if self.claims else self.api_client_id

    @property
    def user_email(self) -> Optional[str]:
        """
        Returns the User Email
        """
        if self.user_data: return self.user_data.email
    

    @property
    def user_email_domain(self) -> Optional[str]:
        """
        Gets the User Email Domain
        """
        if self.user_email is None: return None
        return self.user_email.split('@')[-1].lower()
    
    @property
    def is_system_user(self) -> bool:
        """
        Returns whether the user is a system/api user
        """
        return self.user_type and self.user_type in {
            'api_client', 'service', 'system'
        }

    
    @property
    def is_valid(self) -> bool:
        """
        Returns whether the user is valid
        """
        return self.validation_method is not None


    """
    Extra Properties that can be persisted in the attributes
    """

    @property
    def service_account_name(self) -> Optional[str]:
        """
        Returns the Service Account Name
        """
        if 'service_account_name' not in self.attributes:
            service_account_name = None
            if self.role == UserRole.SYSTEM:
                client_id = self.user_id.replace('@clients', '')
                if service_account_name := self.get_management_api().get_service_client_name(client_id):
                    self.attributes['service_account_name'] = service_account_name.replace('(', '').replace(')', '').replace(' ', '-').lower()
                else:
                    self.attributes['service_account_name'] = client_id
            self.attributes['service_account_name'] = service_account_name
        return self.attributes['service_account_name']

    @property
    def domain_source(self) -> Optional[str]:
        """
        Returns the Domain Source
        """
        if 'domain_source' not in self.attributes:
            self.attributes['domain_source'] = None
        return self.attributes['domain_source']
    
    @property
    def is_local_request(self) -> bool:
        """
        Returns if the request is local
        """
        return self.domain_source and any(
            source in self.domain_source
            for source in {'localhost', '0.0.0.0', '127.0.0.1'}
        )

    @property
    def request_id(self) -> Optional[str]:
        """
        Returns the request id
        """
        return self.data.get('request_id')

    """
    User Roles
    """
    @property
    def user_roles(self) -> Optional[List[str]]:
        """
        Gets the User Roles
        """
        return self.attributes.get('user_roles', [])
    
    @user_roles.setter
    def user_roles(self, value: Optional[Union[str, List[str]]]):
        """
        Sets the User Roles
        """
        if isinstance(value, str): value = [value]
        if not self.attributes.get('user_roles'): self.attributes['user_roles'] = []        
        for role in value:
            role = role.lower()
            if role not in self.attributes['user_roles']:
                self.attributes['user_roles'].append(role)
    
    @user_roles.deleter
    def user_roles(self):
        """
        Deletes the User Roles
        """
        if self.attributes.get('user_roles'):
            _ = self.attributes.pop('user_roles')
    
    

    """
    API Client 
    """

    @property
    def api_client_id(self) -> Optional[str]:
        """
        Gets the API Client ID
        """
        return self.attributes.get('api_client_id')
    
    @api_client_id.setter
    def api_client_id(self, value: Optional[str]):
        """
        Sets the API Client ID
        """
        self.attributes['api_client_id'] = value
    
    @api_client_id.deleter
    def api_client_id(self):
        """
        Deletes the API Client ID
        """
        if self.attributes.get('api_client_id'):
            _ = self.attributes.pop('api_client_id')
    
    @property
    def api_client_identity(self) -> Optional[str]:
        """
        Gets the API Client Identity
        """
        return self.attributes.get('api_client_identity')

    """
    Validators
    """

    def has_user_roles(self, roles: Union[Union[str, int, 'UserRole'], List[Union[str, int, 'UserRole']]], require_all: Optional[bool] = False) -> bool:
        """
        Checks against multiple roles
        """
        if not isinstance(roles, list): roles = [roles]
        valid_roles = [UserRole.parse_role(role) for role in roles]
        if require_all: return all(self.role >= role for role in valid_roles)
        return any(self.role >= role for role in valid_roles)
    
    def has_user_role(self, role: Union[str, int, 'UserRole']) -> bool:
        """
        Checks against a single role
        """
        role = UserRole.parse_role(role)
        return self.role >= role
    
    def has_permission(self, permission: Union[str, int, UserRole]) -> bool:
        """
        Checks if the user has a permission
        """
        permission = UserRole.parse_role(permission)
        return self.role >= permission

    def has_valid_emails(self, emails: List[str]) -> bool:
        """
        Checks if the user has a valid email
        """
        return any(email in self.user_data.email for email in emails)
    
    def has_valid_email_domains(self, domains: List[str]) -> bool:
        """
        Checks if the user has a valid email domain
        """
        return any(domain in self.user_email_domain for domain in domains)
    
    """
    Session Methods
    """
    async def retrieve_and_set_user_session(
        self,
        request: 'Request',
        **kwargs
    ):
        """
        Retrieves the user session and sets it
        """
        session = await self.get_flow_schema('user_session').aget(self, request)
        if session is None: return
        self.session = session
        self.auth_obj.x_api_key = session.api_key
        self.from_session = True
        # logger.info(f'Retrieved User Session: {self.session}')

    async def save_user_session(
        self,
        **kwargs
    ):
        """
        Saves the current session
        """
        if self.session is None: return
        # if self.verbose: logger.info(f'Saving Session: {self.user_id}')
        await self.session_flow.asave(self.session)
        await self.session_flow.asave_data(self.attributes)

    async def delete_user_session(
        self,
        include_data: Optional[bool] = False,
        **kwargs
    ):
        """
        Deletes the current session
        """
        if self.session is None: return
        # if self.verbose: logger.info(f'Deleting Session: {self.user_id}')
        await self.session_flow.adelete()
        if include_data: await self.session_flow.adelete_data()

    async def set_user_session(
        self,
        request: 'Request',
        **kwargs
    ):
        """
        Sets the user session
        """
        if self.session is not None: return
        self.session = self.get_resource('user_session', user_id = self.user_id, api_key = self.api_key)
        await self.save_user_session()


    async def create_or_load_user_session(
        self,
        request: 'Request',
        background_tasks: 'BackgroundTasks' = None, 
        **kwargs
    ):
        """
        Creates or loads the user session
        """
        if self.session is None: 
            self.session = self.get_resource('user_session', user_id = self.user_id, api_key = self.api_key)
            # await self.save_user_session()
        attributes = await self.session_flow.aload_data()
        if attributes: self.update_attributes(attributes)
        if background_tasks: background_tasks.add_task(self.save_user_session)

    
    """
    API Key Data Methods
    """

    async def save_api_key_data(
        self,
        client_id: str,
        user_data: 'AZUserData',
        claims: 'UserJWTClaims',
        **kwargs
    ):
        """
        Persists the API Key Data
        """
        await self.get_flow('api_key', client_id).aset(user_data, claims)

    async def load_api_key_data(
        self,
        client_id: str,
        **kwargs
    ) -> Optional['APIKeyData']:
        """
        Loads the API Key Data
        """
        api_key_data = await self.get_flow('api_key', client_id).afetch()
        if api_key_data is None: return
        self.user_data = api_key_data.user_data
        self.claims = api_key_data.claims


    """
    API Client Methods
    """

    async def configure_api_client_post_authorization(self, request: Optional[Request]):
        """
        Configures the API Client
        """
        api_client_identity = request.headers.get(self.settings.api_client_id_header_key)
        if api_client_identity is None:
            raise InvalidAPIKeyException(detail = f"Invalid API Client Identity: {api_client_identity}")
        if api_client_env := request.headers.get(self.settings.api_client_env_header_key):
            self.attributes['api_client_env'] = api_client_env
        self.attributes['api_client_identity'] = api_client_identity
        if self.role is None: self.role = UserRole.API_CLIENT
        self.api_client_id = f'{self.claims.sub}:{api_client_identity}'
        self.user_type = 'api_client'
        await self.save_api_key_data(client_id = self.api_client_id, user_data = self.user_data, claims = self.claims)
        # await self.save_api_key_data(client_id = api_client_identity, user_data = self.user_data, claims = self.claims)
        self.api_key = self.settings.create_api_key(self.api_client_id, prefix = self.settings.api_client_api_key_prefix)


    """
    Main Auth Methods
    """


    """
    Token Authorizations
    """
    async def validate_authorization(
        self, 
        request: 'Request', 
        token: Optional[str] = None, 
        scopes: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Verifies the Authorization
        """
        self.claims = decode_token(request = request, token = token or self.auth_obj.auth_token, settings = self.settings)
        self.validate_permissions() if scopes is None else self.validate_scopes(scopes)
        try:
            self.user_data = await self.user_data_flow.afetch()
        except Exception as e:
            if '@clients' not in self.claims.sub: 
                self.settings.logger.trace(f'Invalid User Data: {self.claims.sub}', e)
            self.user_data = AZUserData(
                user_id = self.claims.sub, user_metadata = {}, app_metadata = {},
            )
        if self.validation_method is None: self.validation_method = 'auth_token'
        if self.claims.sub.endswith('@clients'):
            if request.headers.get(self.settings.api_client_id_header_key) is not None:
                return await self.configure_api_client_post_authorization(request)
            if self.role is None: self.role = UserRole.SERVICE
            self.user_type = 'service'
        if self.role is None: 
            if self.user_data and self.user_data.user_metadata.get('role'):
                self.role = UserRole.parse_role(self.user_data.user_metadata['role'])
            elif self.user_data and self.user_data.app_metadata.get('role'):
                self.role = UserRole.parse_role(self.user_data.app_metadata['role'])
            else:
                self.role = UserRole.USER
        # Set the api-key
        await self.save_api_key_data(client_id = self.user_id, user_data = self.user_data, claims = self.claims)
        self.api_key = self.settings.create_api_key(self.user_id)


    """
    API Key Authorizations
    """

    async def get_or_create_api_key_data_for_allowed(
        self,
        api_client_id: str,
        **kwargs,
    ):
        """
        Creates an api key data for the allowed client
        """
        api_key_data = await self.get_flow('api_key', api_client_id).afetch()
        if api_key_data is not None:
            self.user_data = api_key_data.user_data
            self.claims = api_key_data.claims
            return
        self.settings.autologger.info(f'Creating API Key Data for {api_client_id}')
        self.user_data = AZUserData(
            user_id = self.api_client_id, 
            user_metadata = {}, 
            app_metadata = {},
        )
        self.claims = self.get_resource(
            'api_key_jwt_claims',
            aud = self.settings.audience,
            azp = self.settings.client_id,
            iss = self.settings.domain,
            sub = self.api_client_id,
            user_roles = [self.role.value],
        )
        await self.save_api_key_data(client_id = api_client_id, user_data = self.user_data, claims = self.claims)
        

    async def validate_allowed_api_key(self, request: Optional[Request], **kwargs):
        """
        Verifies the Allowed API Key
        """
        self.user_type = 'allowed'
        self.api_key = self.auth_obj.x_api_key
        if self.validation_method is None: self.validation_method = 'api_key'
        
        client_config = self.settings.allowed_api_keys.get(self.auth_obj.x_api_key, {})
        if self.role is None: self.role = UserRole.parse_role(client_config.get('role', 'API_CLIENT'))
        
        # Handle setting the other variables / properties
        api_client_identity = request.headers.get(self.settings.api_client_id_header_key, client_config.get('client_name', 'default'))
        self.attributes['api_client_identity'] = api_client_identity
        self.api_client_id = f'allowed.{api_client_identity}'
        await self.get_or_create_api_key_data_for_allowed(self.api_client_id)
        

    async def validate_service_client_api_key(self, request: Optional[Request], **kwargs):
        """
        Verifies the Service Client API Key
        """
        key_hash = get_hashed_key(self.auth_obj.x_api_key)
        if key_hash not in self.settings.api_key_hashes:
            raise InvalidAPIKeyException(detail=f'Invalid Service Client API Key: {self.auth_obj.x_api_key}')
        self.user_type = 'service'
        if self.validation_method is None: self.validation_method = 'api_key'
        if self.role is None: self.role = UserRole.SERVICE
        self.api_key = self.auth_obj.x_api_key
    
    async def refresh_api_client_api_key_data(
        self,
        api_client_id: str,
        request: Optional[Request],
        **kwargs
    ):
        """
        Refreshes the API Client API Key Data
        """
        self.claims = decode_token(request = request, token = self.auth_obj.auth_token)
        self.user_data = AZUserData(
            user_id = self.claims.sub, user_metadata = {}, app_metadata = {},
        )
        await self.save_api_key_data(client_id = api_client_id, user_data = self.user_data, claims = self.claims)
        self.settings.autologger.info(f'Refreshed API Client API Key Data for {api_client_id}')

    async def validate_api_client_api_key(self, request: Optional[Request], scopes: Optional[List[str]] = None, **kwargs):
        """
        Verifies the API Client API Key
        """
        try:
            # xxxxx@clients:client_id
            api_client_id = await self.settings.adecrypt_api_key(self.auth_obj.x_api_key, prefix = self.settings.api_client_api_key_prefix)
            await self.load_api_key_data(client_id = api_client_id)
            # Refresh it if the token is available
            if self.claims is None and self.auth_obj.auth_token:
                await self.refresh_api_client_api_key_data(api_client_id = api_client_id, request = request)
            if self.claims is None:
                raise ExpiredAPIKeyData(f'[API Client] {api_client_id} is expired. Please refresh by providing a valid token.')

            self.validate_permissions() if scopes is None else self.validate_scopes(scopes)
            if self.validation_method is None: self.validation_method = 'api_key'
            if self.role is None: self.role = UserRole.API_CLIENT
            self.user_type = 'api_client'
            self.api_key = self.auth_obj.x_api_key
        except ExpiredAPIKeyData as e:
            raise e
        except Exception as e:
            if self.verbose: logger.trace(f'Invalid API Client API Key: {self.auth_obj.x_api_key}', e)
            else: logger.warning(f'Invalid API Client API Key: {self.auth_obj.x_api_key}: {e}')
            if hasattr(e, 'status_code'): raise e
            raise InvalidAPIKeyException(detail = f'Invalid API Client API Key: {self.auth_obj.x_api_key}') from e

    async def refresh_user_api_key_data(
        self,
        request: Optional[Request],
        **kwargs
    ):
        """
        Refreshes the User API Key Data
        """
        self.claims = decode_token(request = request, token = self.auth_obj.auth_token)
        self.user_data = await self.user_data_flow.afetch()
        await self.save_api_key_data(client_id = self.user_id, user_data = self.user_data, claims = self.claims)
        self.settings.autologger.info(f'Refreshed User API Key Data for {self.user_id}')


    async def validate_user_api_key(self, request: Optional[Request], scopes: Optional[List[str]] = None, **kwargs):
        """
        Verifies the User API Key
        """
        try:
            user_id = await self.settings.adecrypt_api_key(self.auth_obj.x_api_key, prefix = self.settings.api_key_prefix)
            await self.load_api_key_data(client_id = user_id)
            if not self.claims and self.auth_obj.auth_token:
                await self.refresh_user_api_key_data(request = request)
            # if self.claims is None or '@clients' in self.claims.sub:
            if self.claims is None:
                raise ExpiredAPIKeyData(f'[User] {user_id} is expired. Please refresh by reauthenticating.')
            self.validate_permissions() if scopes is None else self.validate_scopes(scopes)
            if self.validation_method is None: self.validation_method = 'api_key'
            if self.role is None: 
                if self.user_data and self.user_data.user_metadata.get('role'):
                    self.role = UserRole.parse_role(self.user_data.user_metadata['role'])
                elif self.user_data and self.user_data.app_metadata.get('role'):
                    self.role = UserRole.parse_role(self.user_data.app_metadata['role'])
                elif '@clients' in self.claims.sub:
                    self.role = UserRole.SERVICE
                else:
                    self.role = UserRole.USER
                # self.role = UserRole.USER
            self.user_type = 'service' if '@clients' in self.claims.sub else 'user'
            self.api_key = self.auth_obj.x_api_key
        except ExpiredAPIKeyData as e:
            raise e
        except Exception as e:
            if self.verbose: logger.trace(f'Invalid User API Key: {self.auth_obj.x_api_key}', e)
            else: logger.warning(f'Invalid User API Key: {self.auth_obj.x_api_key}: {e}')
            if hasattr(e, 'status_code'): raise e
            raise InvalidAPIKeyException(detail = f'Invalid User API Key: {self.auth_obj.x_api_key}') from e
    

    async def validate_api_key(
        self,
        request: 'Request',
        scopes: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Validates the API Key
        """

        # Handle Allowed API Keys
        if self.settings.allowed_api_keys and self.auth_obj.x_api_key in self.settings.allowed_api_keys:
            return await self.validate_allowed_api_key(request = request, scopes = scopes, **kwargs)
        
        # Handle Service Client API Keys
        if not self.auth_obj.x_api_key.startswith(self.settings.api_key_prefix) and (
            self.settings.api_key_suffix and not self.auth_obj.x_api_key.endswith(self.settings.api_key_suffix)
        ):
            return await self.validate_service_client_api_key(request, scopes = scopes, **kwargs)
        
        # Handle API Client API Keys
        if self.auth_obj.x_api_key.startswith(self.settings.api_client_api_key_prefix):
            return await self.validate_api_client_api_key(request, scopes = scopes, **kwargs)

        if not self.auth_obj.x_api_key.startswith(self.settings.api_key_prefix):
            raise InvalidAPIKeyPrefixException(detail = 'Invalid API Key Prefix. Your API Key may be deprecated. Please regenerate your API Key by logging in again.')
        
        # Handle User API Keys
        return await self.validate_user_api_key(request, scopes = scopes, **kwargs)



    """
    Main Validation Methods
    """

    async def run_validation_methods(
        self,
        request: 'Request',
        background_tasks: 'BackgroundTasks' = None,
        **kwargs
    ):
        """
        Runs the validation methods
        """
        # Attempt to validate the API Key first
        # Fallback to the Authorization
        if self.auth_obj.has_x_api_key:
            try:
                return await self.validate_api_key(request = request, **kwargs)
            except Exception as e:
                if not self.auth_obj.has_auth_token: raise e
        return await self.validate_authorization(request = request, **kwargs)
    
    async def run_post_validation_methods(
        self,
        request: 'Request',
        background_tasks: 'BackgroundTasks' = None,
        **kwargs
    ):
        """
        Runs the post validation methods
        """
        await self.run_post_validate_hooks(request)
        if self.is_valid and self.session_enabled: 
            await self.create_or_load_user_session(request, background_tasks = background_tasks, **kwargs)
        
    async def run_validate(
        self, 
        request: 'Request', 
        background_tasks: 'BackgroundTasks' = None, 
        **kwargs
    ):
        """
        This can be subclassed to run the validation
        """
        self.auth_obj = AuthObject.from_request(request, settings = self.settings)
        await self.run_pre_validate_hooks(request)
        if not self.auth_obj.has_auth_token and not self.auth_obj.has_x_api_key:
            raise InvalidAuthorizationException(detail = 'No API Key or Token found')
        
        await self.run_validation_methods(request = request, background_tasks = background_tasks, **kwargs)
        await self.run_post_validation_methods(request, background_tasks = background_tasks, **kwargs)
        
    async def __call__(
        self, 
        request: Request, 
        background_tasks: BackgroundTasks, 
        api_key: APIKey = None, 
        authorization: Authorization = None
    ) -> 'CurrentUser':
        """
        The Middleware Entry Point
        """
        # We use the authorization / api_key params to allow for the 
        # auth to be displayed in the docs
        await self.precall_hook(request)
        try:
            await self.run_validate(request, background_tasks = background_tasks, api_key = api_key, authorization = authorization)
        except AuthZeroException as e:
            if (
                self.verbose
                and 'No API Key or Token found' not in e.detail
                and 'Invalid API Key Prefix' not in e.detail
                and not e.log_error
                and not e.log_devel
            ):
                if self.required:
                    logger.trace('Error Validating User', e)
                else:
                    logger.error(f'Error Validating User: {e}')
            if self.required: raise e
        except Exception as e:
            if self.verbose: 
                if self.required:
                    logger.trace('Error validating user', e)
                    logger.info(self, prefix = self.user_id)
                else:
                    logger.error(f'Error validating user: {e}')
            if self.required: raise e
        return self
    
    async def from_access_token(
        self, 
        access_token: str, 
        request: Optional[Request] = None,
        scope: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        background_tasks: Optional[BackgroundTasks] = None,
        **kwargs,
    ):
        """
        Returns the Current User from the Access Token
        """
        await self.precall_hook(request)
        self.auth_obj = AuthObject(auth_token = access_token)
        await self.run_pre_validate_hooks(request)
        scopes = parse_scopes(scope = scope, scopes = scopes)
        try:
            await self.run_validation_methods(request = request, token = self.auth_obj.auth_token, scopes = scopes, background_tasks = background_tasks, **kwargs)
            await self.run_post_validation_methods(request, background_tasks = background_tasks, **kwargs)
        except Exception as e:
            if self.verbose: 
                if self.required:
                    logger.trace('Error validating user', e)
                    logger.info(self, prefix = self.user_id)
                else:
                    logger.error(f'Error validating user: {e}')
            if self.required: raise e
        return self
    
    async def from_data_sources(
        self, 
        *data: Iterable[Union['Headers', Dict[str, str]]],
        request: Optional[Request] = None,
        scope: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        background_tasks: Optional[BackgroundTasks] = None,
        **kwargs,
    ):
        """
        Returns the Current User from Iterable Data Sources
        """
        await self.precall_hook(request)
        self.auth_obj = AuthObject.from_items(*data, settings = self.settings)
        await self.run_pre_validate_hooks(request)
        if not self.auth_obj.has_auth_token and not self.auth_obj.has_x_api_key:
            if self.verbose: logger.trace('No Auth Token or X-API-Key found')
            if self.required: raise InvalidAuthorizationException()
            return self
        
        scopes = parse_scopes(scope = scope, scopes = scopes)
        try:
            await self.run_validation_methods(request = request, scopes = scopes, **kwargs)
            await self.run_post_validation_methods(request, background_tasks = background_tasks, **kwargs)
        except Exception as e:
            if self.verbose: 
                if self.required:
                    logger.trace('Error validating user', e)
                    logger.info(self, prefix = self.user_id)
                else:
                    logger.error(f'Error validating user: {e}')
            if self.required: raise e
        return self

    
    """
    Hooks
    """

    async def run_pre_validate_hooks(self, request: 'Request', **kwargs):
        """
        Runs the Pre-Validate Hooks
        """
        if self.session_enabled: await self.retrieve_and_set_user_session(request, **kwargs)
        if self._pre_validate_hooks is None: return
        for hook in self._pre_validate_hooks:
            await ThreadPooler.asyncish(hook, self, request)
    
    async def run_post_validate_hooks(self, request: 'Request'):
        """
        Runs the Post-Validate Hooks
        """
        if self.request_id_enabled: 
            await self.configure_request_id_posthook(request)
        if self.domain_source_enabled:
            await self.configure_domain_source_posthook(request)
        if self.role_configure_enabled:
            await self.configure_roles_posthook(request)
        if self._post_validate_hooks is None: return
        for hook in self._post_validate_hooks:
            await ThreadPooler.asyncish(hook, self, request)


    def __getitem__(self, key: str) -> Optional[Any]:
        """
        Returns the user attributes or dict key
        """
        return self.attributes.get(key, getattr(self, key, None))
    
    def __setitem__(self, key: str, value: Any):
        """
        Sets the user attributes or dict key
        """
        if hasattr(self, key): setattr(self, key, value)
        else: self.attributes[key] = value

    def __bool__(self) -> bool:
        """
        Returns True if the user is valid
        """
        return self.is_valid


    def update_attributes(self, attributes: Dict[str, Any]):
        """
        Updates the attributes
        """
        for key, value in attributes.items():
            if isinstance(value, dict):
                self.attributes[key] = self.attributes.get(key, {})
                self.attributes[key].update(value)
            elif isinstance(value, list):
                self.attributes[key] = self.attributes.get(key, [])
                self.attributes[key].extend(value)
                self.attributes[key] = list(set(self.attributes[key]))
            else:
                self.attributes[key] = value
    

    def get_session_cookie_kwargs(self, is_delete: Optional[bool] = None) -> Optional[Dict[str, Any]]:
        """
        Gets the Session Cookie Value
        """
        cookie_kws = {
            "key": self.settings.session_cookie_key,
            "httponly": True,
            "secure": self.settings.is_secure_ingress,
        }
        if not is_delete:
            cookie_kws['value'] = self.session_flow.cache_key
            cookie_kws['expires'] = self.session.expiration_ts
        # Maybe set to expiration now
        return cookie_kws
    

    def get_whoami_data(self, include_data: Optional[bool] = None, **kwargs) -> Dict[str, Any]:
        """
        Returns the Whoami Data
        """
        include_data = include_data if include_data is not None else self.settings.is_development_env
        data = {
            'user': self.user_id, 
            'api_key': self.api_key,
            'request_id': self.request_id,
        }
        if include_data: 
            data['data'] = self.model_dump(mode = 'json', **kwargs)
            if self.session:
                data['session_ttl'] = self.session.ttl
            data['domain_source'] = self.domain_source
        if self.settings.is_development_env and self.session:
            data['session'] = self.session.model_dump(mode = 'json')
        return data

    @classmethod
    def add_post_validate_hook(cls, hook: Callable):
        """
        Adds a post validate hook
        """
        from ..configs.base import AuthZeroContext
        AuthZeroContext.add_post_validate_hook(hook)
    
    @classmethod
    def add_pre_validate_hook(cls, hook: Callable):
        """
        Adds a pre validate hook
        """
        from ..configs.base import AuthZeroContext
        AuthZeroContext.add_pre_validate_hook(hook)

    @classmethod
    def fetch_user_data(cls, user_id: str, **kwargs) -> 'AZUserData':
        """
        Returns the user data
        """
        return cls.get_flow('user_data', user_id = user_id).get(**kwargs)
    
    @classmethod
    async def afetch_user_data(cls, user_id: str, **kwargs) -> 'AZUserData':
        """
        Returns the user data
        """
        return await cls.get_flow('user_data', user_id = user_id).aget(**kwargs)


    """
    Hook Methods
    """

    async def configure_request_id_posthook(self, request: 'Request', **kwargs):
        """
        Configures the request id
        """
        if hasattr(request.state, 'request_id'):
            self.data['request_id'] = request.state.request_id
    
        
    async def configure_domain_source_posthook(
        self,
        request: 'Request',
        **kwargs,
    ):
        """
        Extracts the incoming request url
        """
        headers = {k.lower(): v for k, v in request.headers.items()}
        source = str(headers.get('referer', request.url)).split('?', 1)[0] #  or headers.get('host')
        if source:
            try:
                tld_source = tldextract.extract(source)
                source = f'{tld_source.domain}'
                if tld_source.suffix:
                    source += f'.{tld_source.suffix}'
                if tld_source.subdomain:
                    source = f'{tld_source.subdomain}.{source}'
            except Exception as e:
                # extract the API source from the request
                if ":80" in source or ":443" in source:
                    source = source.rsplit(":", 1)[0]
                if "://" in source:
                    source = source.split("://", 1)[-1]
                if "/" in source:
                    source = source.split("/", 1)[0]
        self.attributes['domain_source'] = source

    
    async def configure_roles_posthook(self, request: 'Request', **kwargs):
        """
        Configures the roles
        """
        if self.role is None: return
        if self.role in [UserRole.API_CLIENT, UserRole.SYSTEM]:
            pass
        elif self.settings.admin_emails and self.user_email and self.user_email in self.settings.admin_emails:
            self.role = UserRole.ADMIN
        elif self.settings.staff_email_domains and self.user_email_domain and self.user_email_domain in self.settings.staff_email_domains:
            self.role = UserRole.STAFF
        roles = [self.role.value]
        if self.claims and self.claims.user_roles:
            roles.extend(self.claims.user_roles)
        self.user_roles = roles



"""
JWT Claims
"""
import time
from pydantic import BaseModel, Field
from typing import Optional, Union, List, Dict


class BaseClaims(BaseModel):
    """
    Base JWT Claims
    """
    aud: Union[str, List[str]]  # API identifier, e.g. "https://app-backend.gex"
    azp: str  # Client app's client_id
    iss: str  # Issuer of the token, always "https://domain.us.auth0.com/"
    scope: str = ''  # scopes requested separated with space, e.g. "models"
    sub: str  # User ID string, e.g. "google-oauth2|..." or "auth0|...."
    cust_user_id_perf_test: Optional[str] = None
    user_roles: Optional[List[str]] = Field(default_factory=list)

    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return False
    
    @property
    def scopes(self) -> List[str]:
        """
        Returns the Scopes
        """
        return self.scope.split(' ')


class APIKeyJWTClaims(BaseClaims):
    """
    API Key JWT Claims
    """

    model_config = {'extra': 'ignore', 'arbitrary_types_allowed': True}


class UserJWTClaims(BaseClaims):
    """
    User JWT Claims
    """
    exp: int  # "expires at" time
    iat: int  # "issued at" time
    
    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return self.exp < (int(time.time()) - 45)

    def to_api_key_claims(self) -> 'APIKeyJWTClaims':
        """
        Converts the User JWT Claims to an API Key JWT Claims
        """
        return APIKeyJWTClaims.model_validate(self, from_attributes=True)



from __future__ import annotations

import gc
import niquests
from urllib.parse import urljoin
from pydantic import BaseModel, Field, ConfigDict
from typing import Optional, List, Dict, Any, Union, Literal, TYPE_CHECKING

"""
PostHog Context Type
{
    "event": "my_event",
    "properties": {
        "$set": ...,
        "my_property": "my_value",
    },
    "distinct_id": "my_distinct_id",
}
"""

PHCtxT = Dict[str, Union[Dict[str, Any], str]]

class PostHogAuth(niquests.auth.BearerTokenAuth):
    """
    The PostHog Auth
    """

    def __init__(self, api_key: str):
        """
        Initializes the PostHog Auth
        """
        self.api_key = api_key
        super().__init__(token = api_key)

    def __call__(self, r: 'niquests.Request') -> 'niquests.Request':
        """
        The Call Method
        """
        r.headers["Authorization"] = f"Bearer {self.api_key}"
        r.headers["Accept"] = "application/json"
        r.headers["Content-Type"] = "application/json"
        return r



class PostHogEndpoint(BaseModel):
    endpoint: str
    
    def get_url(self, *paths: str) -> str:
        """
        Returns the URL
        """
        return urljoin(self.endpoint, '/'.join(paths)).rstrip('/')
    
    @property
    def capture(self) -> str:
        """
        Returns the Capture URL
        """
        return self.get_url('capture')
    
    @property
    def batch(self) -> str:
        """
        Returns the Batch URL
        """
        return self.get_url('batch')
    
    @property
    def identify(self) -> str:
        """
        Returns the Identify URL
        """
        return self.get_url('identify')


class BaseEvent(BaseModel):
    """
    Base Event
    """

    model_config = ConfigDict(extra = 'allow', arbitrary_types_allowed = True)

    @property
    def is_valid(self) -> bool:
        """
        Returns True if the event is valid
        """
        return True 
    

    def prepare_request(
        self,
        exclude_none: Optional[bool] = True,
        batched: Optional[bool] = False,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Formats the event for batching
        """
        return self.model_dump(exclude_none = exclude_none, **kwargs)


class CaptureEvent(BaseEvent):
    """
    The Capture Event
    """
    event: str
    properties: Optional[Dict[str, Any]] = Field(default_factory = dict)
    distinct_id: Optional[str] = None
    timestamp: Optional[int] = None
    token: Optional[str] = None
    context: Optional[Dict[str, Any]] = Field(None)
    
    @property
    def is_valid(self) -> bool:
        """
        Returns True if the event is valid
        """
        return self.event and self.distinct_id
    

    def prepare_request(
        self,
        exclude_none: Optional[bool] = True,
        batched: Optional[bool] = False,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Formats the event for batching
        """
        data = self.model_dump(exclude_none = exclude_none, **kwargs)
        if batched:
            data['properties']['distinct_id'] = data.pop('distinct_id')
            if 'timestamp' in data: data['properties']['timestamp'] = data.pop('timestamp')
        return data



class IdentifyEvent(BaseEvent):
    """
    The Identify Event
    """
    distinct_id: str
    properties: Optional[Dict[str, Any]] = Field(default_factory = dict)
    token: Optional[str] = None
    context: Optional[Dict[str, Any]] = Field(default_factory = dict)
    
    @property
    def is_valid(self) -> bool:
        """
        Returns True if the event is valid
        """
        return self.distinct_id
    

    def prepare_request(
        self,
        exclude_none: Optional[bool] = True,
        batched: Optional[bool] = False,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Formats the event for batching
        """
        data = self.model_dump(exclude_none = exclude_none, **kwargs)
        if batched:
            data['properties']['distinct_id'] = data.pop('distinct_id')
            if 'timestamp' in data: data['properties']['timestamp'] = data.pop('timestamp')
        return data


EventT = Union[CaptureEvent, IdentifyEvent]

class EventQueue(BaseModel):
    """
    The Event Handler
    """
    capture_events: Optional[List[CaptureEvent]] = Field(default_factory = list, description = 'The capture events')
    identify_events: Optional[List[IdentifyEvent]] = Field(default_factory = list, description = 'The identify events')

    def add_event(
        self,
        event: Union[EventT, str],
        **kwargs,
    ):
        """
        Adds an event to the queue
        """
        if isinstance(event, str): event = CaptureEvent(event = event, **kwargs)
        if isinstance(event, CaptureEvent): self.capture_events.append(event)
        elif isinstance(event, IdentifyEvent): self.identify_events.append(event)

    def __len__(self):
        """
        Returns the length of the event queue
        """
        return len(self.capture_events) + len(self.identify_events)
    
    def clear(self):
        """
        Clears the event queue
        """
        self.capture_events.clear()
        self.identify_events.clear()
        gc.collect()
    
    def prepare_events(
        self, 
        kind: Literal['capture', 'identify'] = 'capture',
        batched: Optional[bool] = False,
        exclude_none: Optional[bool] = True,
        clear_after: Optional[bool] = True,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        Prepares the events for serialization
        """
        if kind == 'capture': 
            events = [e.prepare_request(exclude_none = exclude_none, batched = batched, **kwargs) for e in self.capture_events]
            if clear_after: self.capture_events.clear()
            gc.collect()
            return events
            
    def __bool__(self):
        """
        Returns whether the queue is empty
        """
        return len(self) > 0


from __future__ import annotations

"""
Lazy Configurations
"""
from lazyops.utils.lazy import lazy_import
from typing import Any, Dict, Optional, Type, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from .base import AppSettings

_lazy_module_settings: Dict[str, 'AppSettings'] = {}

def register_module_settings(
    module: str,
    settings: 'AppSettings',
):
    """
    Register the module settings
    """
    _lazy_module_settings[module] = settings

def get_module_settings(
    module: Optional[str] = None,
    settings_cls: Optional[Union[Type['AppSettings'], str]] = None,
) -> 'AppSettings':
    """
    Get the module settings
    """
    # Get the last module registered
    if module is None: 
        if not _lazy_module_settings:
            raise ValueError("No module settings registered")
        module = list(_lazy_module_settings.keys())[-1]
    if module not in _lazy_module_settings:
        if settings_cls is None:
            try:
                settings_cls = lazy_import(f'{module}.configs.settings')
            except ImportError as e:
                raise ValueError(f"Module settings for {module} not found") from e
        elif isinstance(settings_cls, str):
            settings_cls = lazy_import(settings_cls)
        if isinstance(settings_cls, type):
            settings_cls = settings_cls()
        _lazy_module_settings[module] = settings_cls
    return _lazy_module_settings[module]




from __future__ import annotations


import datetime
from typing import Any, Dict, List, Optional, Union


def remove_sql_comments(sql_text: str) -> str:
    """
    Removes the SQL comments from the given SQL text
    """
    while "/*" in sql_text:
        sql_text = sql_text[:sql_text.index("/*")] + sql_text[sql_text.index("*/") + 2:].strip()
    # if "--" in sql_text:
    return sql_text


def build_sql_metadata_filter(
    conditional: Optional[str] = 'AND',
    metadata_key: Optional[str] = '_metadata',
    **filters: Dict[str, Union[int, float, datetime.datetime, Dict, List, Any]]
) -> str:
    """
    Constructs the WHERE clause for the `_metadata` property because it's a `jsonb` field
    """

    q = ""
    for key, value in filters.items():
        if 'date' in key:
            # Handle dates
            if isinstance(value, list):
                # Handle date ranges
                if isinstance(value[0], datetime.datetime):
                    value[0] = value[0].strftime("%Y-%m-%d %H:%M:%S")
                    value[1] = value[1].strftime("%Y-%m-%d %H:%M:%S")
                q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ BETWEEN {value[0]} AND {value[1]} || @ == null)' {conditional} "
            else:
                # Handle single dates
                op = '>' if key == "open_date" else '<='
                q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ {op} {value} || @ == null)' {conditional} "
            continue

        # Handle Ints and Floats
        if isinstance(value, (int, float)):
            op = '>=' if 'min' in key else '<='
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ {op} {value} || @ == null)' {conditional} "
            continue

        # Handle Strings
        if isinstance(value, str):
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ ILIKE %'{value}'% || @ == null)' {conditional} "
            continue

        # Handle Lists
        if isinstance(value, list):
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ && {value} || @ == null)' {conditional} "
            continue

        # Handle Dicts
        if isinstance(value, dict):
            q += f"jsonb_path_exists({metadata_key}, '$.{key} ? (@ ?& {value.keys()} || @ == null)' {conditional} "
            continue

    q = q[:-len(conditional)-1]
    return q


from __future__ import annotations

"""
Generic Errors
"""

import traceback
from lazyops.utils.logs import logger
from fastapi.exceptions import HTTPException
from typing import Any, Optional, Union


_is_development_env: Optional[bool] = None

def set_development_env(is_development: bool):
    """
    Sets the development environment
    """
    global _is_development_env
    _is_development_env = is_development


class AppException(Exception):

    base: Optional[str] = None
    concat_detail: Optional[bool] = True
    verbosity: Optional[int] = None

    log_error: Optional[bool] = False # Always log the error message
    log_devel: Optional[bool] = False # Log the error message if in development mode

    """
    Base App Exception
    """

    def __init__(
        self, 
        detail: str = None, 
        error: Optional[Exception] = None, 
        **kwargs
    ):
        """
        Constructor
        """
        message = ""
        if not detail or self.concat_detail:
            message += f"{self.base}: " or ''
        if detail: message += detail
        if error and self.verbosity is not None:
            if self.verbosity >= 1:
                message += f"\nError: {error}"
            elif self.verbosity <= 5:
                message += f"\nError: {error}\nTraceback: {traceback.format_exc()}"
        super().__init__(message, **kwargs)
        if self.log_error or (self.log_devel and self.is_development_env):
            self.display(message)

    @property
    def is_development_env(self) -> bool:
        """
        Returns True if the app environment is development
        """
        return _is_development_env
    
    def display(self, message: str):
        """
        Displays the error
        """
        logger.error(f'[{self.__class__.__name__}]: {message}')

class DatabaseItemNotFoundException(AppException):
    """
    Database Item Not Found Exception
    """
    base = "Database Item Not Found"
    verbosity = 1
    log_devel = True

try:
    from fastapi.exceptions import HTTPException

    class APIException(HTTPException):
        """
        Base API Exception
        """
        
        base: Optional[str] = None
        concat_detail: Optional[bool] = True
        verbosity: Optional[int] = None
        default_status_code: Optional[int] = 500

        log_error: Optional[bool] = False # Always log the error message
        log_devel: Optional[bool] = False # Log the error message if in development mode

        def __init__(
            self, 
            detail: str = None, 
            error: Optional[Exception] = None, 
            status_code: Optional[int] = None, 
            **kwargs
        ):
            """
            Constructor
            """
            message = ""
            if not detail or self.concat_detail:
                message += f"{self.base}: " or ''
            if detail: message += detail
            if error and self.verbosity is not None:
                if self.verbosity >= 1:
                    message += f"\nError: {error}"
                elif self.verbosity <= 5:
                    message += f"\nError: {error}\nTraceback: {traceback.format_exc()}"
            status_code = status_code or self.default_status_code
            super().__init__(status_code=status_code, detail=message, **kwargs)
            if self.log_error or (self.log_devel and self.is_development_env):
                self.display(message, status_code)

        @property
        def is_development_env(self) -> bool:
            """
            Returns True if the app environment is development
            """
            return _is_development_env
        
        def display(self, message: str, status_code: int):
            """
            Displays the error
            """
            logger.error(f'[{self.__class__.__name__} - {status_code}]: {message}')
    

    class DatabaseItemNotFoundException(APIException):
        """
        Database Item Not Found Exception
        """
        base = "Database Item Not Found"
        verbosity = 1
        log_devel = True
        default_status_code = 404
    
    

except ImportError:
    pass



"""
The SQLite Model Registry
"""

import functools
from pydantic import BaseModel, computed_field
from typing import Any, Dict, List, Optional, Type, Union, TYPE_CHECKING

if TYPE_CHECKING:
    import sqlite3
    # import aiosqlite
    from lazyops.libs import aiosqlite
    from .mixins import SQLiteModelMixin


class SQLiteModelConfig(BaseModel):
    """
    The SQLite Model Schema
    """
    tablename: str
    sql_fields: Dict[str, Union[str, List[str], Dict[str, str]]]
    sql_pkey: str
    sql_keys: List[str]
    sql_insert: str
    sql_insert_q: str
    search_precisions: Dict[str, int]
    autoset: Optional[bool] = None
    sql_model_name: Optional[str] = None
    
    db_conn: Optional[Any] = None
    
    if TYPE_CHECKING:
        db_conn: Union['sqlite3.Connection', 'aiosqlite.Connection'] = None
    

    @property
    def conn(self) -> Union['sqlite3.Connection', 'aiosqlite.Connection']:
        """
        Returns the connection
        """
        if self.db_conn is None:
            if self.sql_model_name in _sqlite_model_name_to_connection:
                self.db_conn = _sqlite_model_name_to_connection[self.sql_model_name]
            else:
                raise ValueError(f'Model {self.sql_model_name} not registered and no connection provided')
        return self.db_conn

    def __getitem__(self, key: str) -> Any:
        """
        Gets the item
        """
        return getattr(self, key)
    
    @computed_field
    @property
    def sql_schema(self) -> Dict[str, Union[str, List[str], Dict[str, str]]]:
        """
        Returns the SQLite Schema
        """
        return {
            'tablename': self.tablename,
            'sql_fields': self.sql_fields,
            'sql_pkey': self.sql_pkey,
            'sql_keys': self.sql_keys,
            'sql_insert': self.sql_insert,
            'sql_insert_q': self.sql_insert_q,
            'search_precisions': self.search_precisions,
            'autoset': self.autoset,
        }
    

_sqlite_model_schema_registry: Dict[str, Dict[str, Union[str, List[str], Dict[str, str]]]] = {}
_sqlite_model_name_to_tablename: Dict[str, str] = {}
_sqlite_model_name_to_connection: Dict[str, Union['sqlite3.Connection', 'aiosqlite.Connection']] = {}
_sqlite_model_config_registry: Dict[str, SQLiteModelConfig] = {}


def get_or_register_sqlite_tablename(
    model: 'SQLiteModelMixin',
    tablename: Optional[str] = None,
) -> str:
    """
    Registers the SQLite Tablename
    """
    global _sqlite_model_name_to_tablename
    model_name = f'{model.__module__}.{model.__name__}'
    if not tablename and model_name not in _sqlite_model_name_to_tablename:
        if hasattr(model, 'tablename') and model.model_fields['tablename'].default is not None:
            tablename = model.model_fields['tablename'].default
        else:
            raise ValueError(f'Model {model_name} not registered and no tablename provided')
    if tablename and model_name not in _sqlite_model_name_to_tablename:
        _sqlite_model_name_to_tablename[model_name] = tablename
    
    if model_name not in _sqlite_model_name_to_tablename:
        raise ValueError(f'Model {model_name} not registered and no tablename provided')
    return _sqlite_model_name_to_tablename[model_name]

def get_or_register_sqlite_connection(
    model: 'SQLiteModelMixin',
    conn: Optional[Union['sqlite3.Connection', 'aiosqlite.Connection']] = None,
) -> Union['sqlite3.Connection', 'aiosqlite.Connection']:
    """
    Registers the SQLite Connection
    """
    global _sqlite_model_name_to_connection
    model_name = f'{model.__module__}.{model.__name__}'
    if conn and model_name not in _sqlite_model_name_to_connection:
        _sqlite_model_name_to_connection[model_name] = conn
    
    if model_name not in _sqlite_model_name_to_connection:
        raise ValueError(f'Model {model_name} not registered and no connection provided')
    return _sqlite_model_name_to_connection[model_name]

def get_or_register_sqlite_schema(
    model: 'SQLiteModelMixin',
    tablename: Optional[str] = None,
    auto_set: Optional[bool] = None,
    conn: Optional[Union['sqlite3.Connection', 'aiosqlite.Connection']] = None,
) -> Dict[str, Union[str, List[str], Dict[str, Union[str, int]]]]:
    # sourcery skip: extract-method
    """
    Registers the SQLite Schema
    """
    global _sqlite_model_schema_registry
    tablename = get_or_register_sqlite_tablename(model, tablename)
    if tablename not in _sqlite_model_schema_registry:
        sql_fields, sql_pkey, search_precisions = model._get_sql_field_schema()
        sql_keys = list(sql_fields.keys())
        sql_insert_q = ('?, ' * len(sql_keys)).rstrip(', ')
        sql_insert = ', '.join(sql_keys)
        _sqlite_model_schema_registry[tablename] = {
            'tablename': tablename,
            'sql_fields': sql_fields,
            'sql_pkey': sql_pkey,
            'sql_keys': sql_keys,
            'sql_insert': sql_insert,
            'sql_insert_q': sql_insert_q,
            'search_precisions': search_precisions,
        }
        if conn is not None: get_or_register_sqlite_connection(model, conn)
    if auto_set is not None: _sqlite_model_schema_registry[tablename]['auto_set'] = auto_set
    return _sqlite_model_schema_registry[tablename]


def get_sqlite_model_pkey(
    model: 'SQLiteModelMixin',
    tablename: Optional[str] = None,
) -> str:
    """
    Returns the SQLite Model PKey
    """
    tablename = get_or_register_sqlite_tablename(model, tablename)
    return _sqlite_model_schema_registry[tablename]['sql_pkey']
    


def has_auto_set_in_schema(
    model: 'SQLiteModelMixin',
) -> bool:
    """
    Checks if the model has auto set in the schema
    """
    return 'auto_set' in get_or_register_sqlite_schema(model)

@functools.lru_cache()
def retrieve_sqlite_model_schema(
    model_name: str
) -> Dict[str, Union[str, List[str], Dict[str, Union[str, int]]]]:
    """
    Retrieves the SQLite Model Schema
    """
    if model_name not in _sqlite_model_name_to_tablename:
        raise ValueError(f'Model {model_name} not registered')
    tablename = _sqlite_model_name_to_tablename[model_name]
    if tablename not in _sqlite_model_schema_registry:
        raise ValueError(f'Model {model_name} not registered')
    return _sqlite_model_schema_registry[tablename]


def get_sqlite_model_config(
    model: 'SQLiteModelMixin',
    tablename: Optional[str] = None,
) -> SQLiteModelConfig:
    """
    Returns the SQLite Model Config
    """
    model_name = f'{model.__module__}.{model.__name__}'
    if model_name not in _sqlite_model_config_registry:
        if tablename and tablename in _sqlite_model_schema_registry:
            schema = _sqlite_model_schema_registry[tablename]
        else:
            schema = retrieve_sqlite_model_schema(model_name)
        config = SQLiteModelConfig.model_validate(schema)
        config.sql_model_name = model_name
        if model_name in _sqlite_model_name_to_connection:
            config.db_conn = _sqlite_model_name_to_connection[model_name]
        _sqlite_model_config_registry[model_name] = config
    return _sqlite_model_config_registry[model_name]


import re
import html
import contextlib
from typing import Optional

html_re_pattern = re.compile('<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});')
underscores = re.compile("___+")
name_pattern_regex = re.compile(r'(?:(?<=^)|(?<=[^A-Za-z.,]))[A-Za-z.,]+(?: [A-Za-z.,]+)*(?:(?=[^A-Za-z.,])|(?=$))')


def cleanup_whitespace(text: str) -> str:
    """
    Cleans up the whitespace
    """
    return re.sub(r'\s+', ' ', text).strip()

def cleanup_dots(text: str) -> str:
    """
    Cleans up the dots found in table of contents

    INTRODUCTION............................................ > INTRODUCTION.
    """
    return re.sub(r'\.+', '.', text).strip()


def extract_email(text: str) -> Optional[str]:
    """
    Extracts the email
    """
    match = re.search(r'[\w\.-]+@[\w\.-]+', text)
    return match[0] if match else None


def extract_phone(text: str) -> Optional[str]:
    """
    Extracts the phone
    """
    # 205460001
    # 3014430995
    match = re.search(r'[\d\(\)\-]+', text)
    if not match: return None
    pn = match[0]
    if len(pn.replace('(', '').replace(')', '').replace('-', '').replace(' ', '')) <= 9: return None
    return pn


def unescape_html(text: str, replace_breaks: Optional[bool] = True) -> str:
    """
    Unescape HTML
    """
    text = html.unescape(text)
    if replace_breaks: text = text.replace('<br/>', '\n')
    return text.strip()


def clean_html(text: str) -> str:
    """
    Clean HTML
    """
    return html_re_pattern.sub('', html.unescape(text))


def clean_text(text: str) -> str:
    """
    Cleans the text
    """
    if '�' in text: text = text.replace('�', '')
    
    # Find and remove double underscores
    if underscores.search(text):
        text = underscores.sub(' ', text)
    
    # Cleanup HTML
    if '<' in text and '>' in text: 
        with contextlib.suppress(Exception):
            from markdownify import markdownify
            text = markdownify(text)
            # text = cleanup_whitespace(text)
    
    return cleanup_whitespace(text)
    # return text


from __future__ import annotations

"""
AuthZero Clients
"""
from .base import BaseModel
from pydantic import Field, PrivateAttr
from ..utils.lazy import logger
from typing import Optional, Dict, Any, Union, List, TYPE_CHECKING

class AuthZeroClientObject(BaseModel):
    """
    The Auth Zero Client Object
    """
    
    client_id: str
    tenant: str
    name: str
    client_secret: str
    description: Optional[str] = None
    is_global: bool = Field(False, description = 'If True, the client is global', validation_alias = 'global')
    app_type: Optional[str] = None
    logo_uri: Optional[str] = None
    is_first_party: bool = Field(False, description = 'If True, the client is a first party client')
    oidc_conformant: bool = Field(False, description = 'If True, the client is an OIDC conformant client')
    callbacks: Optional[List[str]] = Field(default_factory = list, description = 'The callbacks')
    allowed_origins: Optional[List[str]] = Field(default_factory = list, description = 'The allowed origins')
    web_origins: Optional[List[str]] = Field(default_factory = list, description = 'The web origins')
    client_aliases: Optional[List[str]] = None
    allowed_clients: Optional[List[str]] = None
    allowed_logout_urls: Optional[List[str]] = Field(default_factory = list, description = 'The allowed logout urls')
    oidc_logout: Optional[Dict[str, Any]] = None
    grant_types: Optional[List[str]] = None
    jwt_configuration: Optional[Dict[str, Any]] = None
    signing_keys: Optional[List[Dict[str, Any]]] = None
    encryption_key: Optional[Dict[str, Any]] = None
    sso: Optional[bool] = None
    sso_disabled: Optional[bool] = None
    cross_origin_authentification: Optional[bool] = None
    cross_origin_loc: Optional[str] = None
    custom_login_page_on: Optional[bool] = None
    custom_login_page: Optional[str] = None
    cuustom_login_page_preview: Optional[str] = None
    form_template: Optional[str] = None
    addons: Optional[Dict[str, Any]] = None
    token_endpoint_auth_method: Optional[str] = None
    client_metadata: Optional[Dict[str, Any]] = None
    mobile: Optional[Dict[str, Any]] = None
    initiate_login_uri: Optional[str] = None
    native_social_login: Optional[Dict[str, Any]] = None
    refresh_token: Optional[Dict[str, Any]] = None
    organization_usage: Optional[str] = None
    organization_require_behavior: Optional[str] = None
    client_authentication_methods: Optional[Dict[str, Any]] = None
    require_pushed_authorization_requests: Optional[bool] = None
    access_token: Optional[Dict[str, Any]] = None
    signed_request_object: Optional[Dict[str, Any]] = None
    compliance_level: Optional[str] = None

    model_config = {'arbitrary_types_allowed': True, 'extra': 'allow'}

    # Extra Private Attributes
    _needs_update: Optional[bool] = PrivateAttr(False)

    def get_app_update_counts(self) -> Dict[str, int]:
        """
        Returns the update counts
        """
        return {
            'allowed_origins': len(self.allowed_origins),
            'callbacks': len(self.callbacks),
            'web_origins': len(self.web_origins),
            'allowed_logout_urls': len(self.allowed_logout_urls),
        }
    
    def get_app_patch_data(self) -> Dict[str, List[str]]:
        """
        Returns the patch data
        """
        return {
            'allowed_origins': self.allowed_origins,
            'callbacks': self.callbacks,
            'web_origins': self.web_origins,
            'allowed_logout_urls': self.allowed_logout_urls,
        }

    def add_app_url(
        self,
        allowed_origin: Optional[str] = None,
        callback: Optional[str] = None,
        web_origin: Optional[str] = None,
        allowed_logout_url: Optional[str] = None,
        verbose: Optional[bool] = None,
    ):
        """
        Adds to app urls
        """

        if allowed_origin is not None and allowed_origin not in self.allowed_origins:
            self.allowed_origins.append(allowed_origin)
            self._needs_update = True
            if verbose: logger.info(f'Added allowed origin: {allowed_origin}')
        if callback is not None and callback not in self.callbacks:
            self.callbacks.append(callback)
            self._needs_update = True
            if verbose: logger.info(f'Added callback: {callback}')
        if web_origin is not None and web_origin not in self.web_origins:
            self.web_origins.append(web_origin)
            self._needs_update = True
            if verbose: logger.info(f'Added web origin: {web_origin}')
        if allowed_logout_url is not None and allowed_logout_url not in self.allowed_logout_urls:
            self.allowed_logout_urls.append(allowed_logout_url)
            self._needs_update = True
            if verbose: logger.info(f'Added allowed logout url: {allowed_logout_url}')
    



from __future__ import annotations

"""
Base ORM Object Types
"""

# Note - The original errors for the dataclass implementation
# was due to pydantic's dataclass implementation not being
# compatible with sqlalchemy's mapped dataclass implementation

# https://docs.sqlalchemy.org/en/20/orm/extensions/asyncio.html#preventing-implicit-io-when-using-asyncsession
# https://github.com/sqlalchemy/sqlalchemy/discussions/7336

import datetime

from pydantic import BaseModel
from pydantic.alias_generators import to_snake
from dataclasses import dataclass
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import as_declarative, declared_attr
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import Mapped
from sqlalchemy import MetaData
from sqlalchemy.orm import mapped_column
from sqlalchemy import Text, Table
from sqlalchemy import func as sql_func
# from sqlalchemy.orm import Mapped, Relationship, relationship
from sqlalchemy.orm import defer
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect
from sqlalchemy.sql.expression import text, select, Select, ColumnElement, and_, update, Update, delete, or_
from sqlalchemy.dialects.postgresql import Insert, insert
from sqlalchemy.orm import selectinload
from sqlalchemy.orm import registry

from ...types import errors
from typing import Optional, Type, TypeVar, Union, Set, Any, Tuple, List, Dict, cast, Generic, Generator, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from lazyops.utils.logs import Logger
    from lazyops.libs.pooler import ThreadPool

mapper_registry = registry()


class Base(AsyncAttrs, DeclarativeBase):
    """
    A Base ORM Object
    """

    if TYPE_CHECKING:
        id: Mapped[Any]
        created_at: Mapped[datetime.datetime]
        updated_at: Mapped[datetime.datetime]


class BasePydantic(MappedAsDataclass, DeclarativeBase, dataclass_callable = dataclass):
    """
    Base Pydantic ORM Object
    """

    if TYPE_CHECKING:
        id: Mapped[Any]
        created_at: Mapped[datetime.datetime]
        updated_at: Mapped[datetime.datetime]

    
    def __init__(self, **kw):
        mapper = inspect(self).mapper
        for key in mapper.relationships:
            if key in kw:
                kw[key] = mapper.relationships[key].entity.class_(**kw[key])
        super().__init__(**kw)

    @property
    def orm_class_name(self) -> str:
        """
        Returns the ORM class name
        """
        return self.__class__.__name__
    
    @property
    def orm_parent_class_name(self) -> str:
        """
        Returns the ORM parent class name
        """
        return self.__class__.__qualname__.split('.')[3]

    def get_non_relationship_fields(
        self,
        include: Optional[Set[str]] = None,
        exclude: Optional[Set[str]] = None,
        **kwargs
    ) -> List[str]:
        """
        Returns the non relationship fields
        """
        include = include or set()
        exclude = exclude or set()
        return [
            k for k in self.__dict__ if (
                k not in self.__mapper__.relationships or \
                k in include or \
                k not in exclude
            )
        ]
    
    def get_relationship_fields(
        self,
        include: Optional[Set[str]] = None,
        exclude: Optional[Set[str]] = None,
        **kwargs
    ) -> List[str]:
        """
        Returns the relationship fields
        """
        include = include or set()
        exclude = exclude or set()
        return [
            k for k in self.__mapper__.relationships if (
                k not in include or \
                k in exclude
            )
        ]

    def get_exportable_kwargs(
        self,
        include: Any = None,
        exclude: Any = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Returns the exportable kwargs
        """
        data = {
            k: v for k, v in self.__dict__.items() if k in self.get_non_relationship_fields(
                include = include,
                exclude = exclude,
                **kwargs,
            )
        }
        if exclude_none:
            data = {k: v for k, v in data.items() if v is not None}
        if exclude_unset or exclude_defaults:
            data = {k: v for k, v in data.items() if v != self.__mapper__.columns[k].default.arg}
        return data

class BaseORMModel(BaseModel):
    """
    Generic Base ORM Model for Pydantic
    using this to allow for type checking
    """
    if TYPE_CHECKING:
        id: int
        created_at: datetime.datetime
        updated_at: datetime.datetime
    

@as_declarative()
class BaseORM:
    """
    Base ORM Object
    """
    if TYPE_CHECKING:
        id: Mapped[Any]
        created_at: Mapped[datetime.datetime]
        updated_at: Mapped[datetime.datetime]


    __name__: str


ModelTypeBaseORM = TypeVar("ModelTypeBaseORM", bound = BaseORM)
ModelTypeBasePydantic = TypeVar("ModelTypeBasePydantic", bound = BasePydantic)
ModelType = Union[ModelTypeBasePydantic, ModelTypeBaseORM]

SourceSchemaType = TypeVar("SourceSchemaType", bound = BaseORMModel)
ResultObject = TypeVar("ResultObject", BaseORMModel, BasePydantic)

async def has_this_obj(db: AsyncSession, obj: ModelType, attr: str, obj_id: Union[int, str]) -> bool:
    """
    Returns whether or not the object has this object
    """
    await db.refresh(obj, [attr])
    return any(item.id == obj_id for item in getattr(obj, attr))


async def append_obj(
    db: AsyncSession,
    *,
    parent_obj: ModelType,
    child_obj: ModelType,
    attr: str
) -> Tuple[ModelType, ModelType]:
    """
    Appends the object to the parent object
    """
    await db.refresh(parent_obj, [attr])
    items: List[ModelType] = getattr(parent_obj, attr)
    items.append(child_obj)
    setattr(parent_obj, attr, items)

    await db.commit()
    await db.refresh(parent_obj)
    await db.refresh(child_obj)
    return parent_obj, child_obj


async def remove_obj(
    db: AsyncSession,
    *,
    parent_obj: ModelType,
    child_obj: ModelType,
    attr: str
) -> Tuple[ModelType, ModelType]:
    """
    Removes the object from the parent object
    """
    await db.refresh(parent_obj, [attr])
    items: List[ModelType] = getattr(parent_obj, attr)
    items.remove(child_obj)
    setattr(parent_obj, attr, items)

    await db.commit()
    await db.refresh(parent_obj)
    await db.refresh(child_obj)
    return parent_obj, child_obj


def prepare_update_data(
    model: Type[ModelType], 
    update_data: Union[SourceSchemaType, Dict[str, Any]]
) -> Dict[str, Any]:
    """
    Prepares the update data
    """
    if isinstance(update_data, BaseModel):
        update_data = update_data.model_dump(exclude_unset=True)
    update_data = {k: v for k, v in update_data.items() if hasattr(model, k)}
    if hasattr(model, 'updated_at'):
        update_data['updated_at'] = datetime.datetime.now(datetime.timezone.utc)
    return update_data


"""
Credits to https://github.com/dmitrybabanovforreal
"""

class ObjectCRUD(Generic[ModelTypeBasePydantic, SourceSchemaType]):

    def __init__(
        self, 
        model: Type[Union[BaseORMModel, BasePydantic]], 
        schema: Optional[Type[BaseORMModel]] = None,
        defer_attrs: List[str] = None,
        auto_commit: Optional[bool] = True,
        **kwargs,
    ):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).

        **Parameters**

        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic schema class
        * `defer_attrs`: A list of attributes to defer loading
        * `auto_commit`: Whether or not to auto commit (True by default. Use False to allow for batch operations)
        """
        self.model = model
        
        self.defer_attrs = defer_attrs or []
        self.auto_commit = auto_commit
        self._schema = schema
        self._logger: Optional['Logger'] = None
        self._pooler: Optional['ThreadPool'] = None
        self._kwargs = kwargs
        self._lazyattrs: Dict[str, Any] = {}
        self._post_init(**kwargs)

    def _post_init(self, **kwargs):
        """
        Post-Init
        """
        pass


    @property
    def table_name(self) -> str:
        """
        Returns the table name
        """
        if 'table_name' not in self._lazyattrs:
            self._lazyattrs['table_name'] = self.model.__tablename__
        return self._lazyattrs['table_name']
    
    @property
    def table_schema_name(self) -> str:
        """
        Returns the table schema name
        """
        if 'table_schema_name' not in self._lazyattrs:
            self._lazyattrs['table_schema_name'] = self.model.metadata.schema
        return self._lazyattrs['table_schema_name']
    
    @property
    def table_with_schema_name(self) -> str:
        """
        Returns the table with schema name

        {table_schema_name}.{table_name}
        """
        if 'table_with_schema_name' not in self._lazyattrs:
            if self.table_schema_name:
                self._lazyattrs['table_with_schema_name'] = f'{self.table_schema_name}.{self.table_name}'
            else:
                self._lazyattrs['table_with_schema_name'] = self.table_name
        return self._lazyattrs['table_with_schema_name']
    
    
    @property
    def pooler(self) -> 'ThreadPool':
        """
        Returns the pooler
        """
        if self._pooler is None:
            from lazyops.libs.pooler import ThreadPooler
            self._pooler = ThreadPooler
        return self._pooler
    
    @property
    def table(self) -> Table:
        """
        Returns the table
        """
        return self.model.__table__
    

    @property
    def object_class_name(self) -> str:
        """
        Returns the object class name
        """
        if 'object_class_name' not in self._lazyattrs:
            self._lazyattrs['object_class_name'] = self.model.__name__
        return self._lazyattrs['object_class_name']

    @property
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        if not self._logger:
            from lazyops.utils.logs import logger
            self._logger = logger
        return self._logger
    
    """
    Getters
    """
    
    async def all(self, db: AsyncSession, stmt: Select) -> List[ResultObject]:
        """
        Returns all objects of the query
        """
        res = (await db.scalars(stmt)).all()
        res = cast(List[ModelType], res)
        return res
    
    async def first(
        self, 
        db: AsyncSession, 
        stmt: Union[Select, Insert, Update], 
        raise_for_not_found: bool = False
    ) -> ResultObject:
        """
        Returns the first object of the query
        """
        stmt = stmt.limit(1) # Potential cause for memory leak
        res = (await db.scalars(stmt)).first()
        res = cast(self.model, res)
        self._raise_for_not_found(res, raise_for_not_found)
        return res


    async def get(
        self,
        db: AsyncSession,
        id: Any,
        raise_for_not_found: bool = False,
        defer_attrs: List[str] = None,
    ) -> Optional[ResultObject]:
        """
        Returns the object with the given id
        """
        stmt = select(self.model).where(self.model.id == id)
        if defer_attrs: stmt = self._defer_attrs(stmt, defer_attrs)
        return await self.first(db, stmt, raise_for_not_found)

    async def get_first(
        self, 
        db: AsyncSession, 
        raise_for_not_found: bool = False, 
        **filters
    ) -> Optional[ResultObject]:
        """
        Gets the first object with the given filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(self.model).where(and_stmt)
        stmt = self._defer_attrs(stmt, None)
        return await self.first(db, stmt, raise_for_not_found)


    async def get_many(
        self, 
        db: AsyncSession, *, 
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        defer_attrs: List[str] = None,
    ) -> List[ResultObject]:
        """
        Gets multiple objects with pagination
        """
        stmt = select(self.model)
        if skip is not None: stmt = stmt.offset(skip)
        if limit is not None: stmt = stmt.limit(limit)
        stmt = self._defer_attrs(stmt, defer_attrs)
        return await self.all(db, stmt)

    async def get_many_filtered(
        self, 
        db: AsyncSession, 
        *, 
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        **filters
    ) -> List[ResultObject]:
        """
        Gets multiple objects with pagination and filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(self.model).where(and_stmt)
        if skip is not None: stmt = stmt.offset(skip)
        if limit is not None: stmt = stmt.limit(limit)
        stmt = self._defer_attrs(stmt, None)
        return await self.all(db, stmt)
    
    async def get_many_by_ids(
        self,
        db: AsyncSession,
        *,
        ids: List,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        load_children: List[str] = None,
        raise_for_not_found: bool = False
    ) -> List[SourceSchemaType]:
        """
        Gets multiple objects with pagination and filters by ids
        """
        ids_string = ', '.join([f"'{i}'" for i in ids])
        stmt = select(self.model).where(text(f'id in ({ids_string})'))
        if skip is not None: stmt = stmt.offset(skip)
        if limit is not None: stmt = stmt.limit(limit)
        if load_children: stmt = self._add_related_obj_loading(stmt, load_children)
        objs = await self.all(db, stmt)
        if raise_for_not_found and len(objs) < len(ids):
            self._raise_for_not_found(None, raise_for_not_found=True)
        return objs
    

    async def get_random(
        self, 
        db: AsyncSession, 
        *, 
        limit: int = 100, 
        **filters
    ) -> List[ResultObject]:
        """
        Gets multiple objects with pagination and filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(self.model).where(and_stmt).order_by(sql_func.random()).limit(limit)
        stmt = self._defer_attrs(stmt, None)
        return await self.all(db, stmt)
    
    """
    Count Operations
    """

    async def get_count(
        self, 
        db: AsyncSession, 
        **filters
    ) -> int:
        """
        Gets the count of objects with filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = select(sql_func.count(self.model.id)).where(and_stmt)
        return await db.scalar(stmt)



    """
    Create Operations
    """

    def prepare_encoded_object(
        self,
        obj_in: SourceSchemaType,
        method: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Encodes the object
        """
        return jsonable_encoder(obj_in, **kwargs)
    
    async def prepare_encoded_object_one(
        self,
        obj_in: SourceSchemaType,
        method: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Encodes the object
        """
        values = await self.pooler.arun(self.prepare_encoded_object, obj_in, method, **kwargs)
        if method:
            if method == 'upsert' and hasattr(self.model, 'updated_at'):
                values['updated_at'] = datetime.datetime.now(datetime.timezone.utc)
            elif method == 'create' and hasattr(self.model, 'created_at'):
                values['created_at'] = datetime.datetime.now(datetime.timezone.utc)
        return values


    async def batch_prepare_encoded_objects(
        self,
        objs_in: List[SourceSchemaType],
        method: str,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        Encodes the object
        """
        return await self.pooler.async_map(
            self.prepare_encoded_object_one,
            objs_in,
            method = method,
            **kwargs
        )

    async def create(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType
    ) -> ModelType:
        """
        Creates an object
        """
        obj_in_data = self.prepare_encoded_object(obj_in, method = 'create')
        db_obj = self.model(**obj_in_data)  # type: ignore
        if hasattr(db_obj, 'created_at'):
            db_obj.created_at = datetime.datetime.now(datetime.timezone.utc)
        db.add(db_obj)
        if self.auto_commit:
            await db.commit()
            await db.refresh(db_obj)
        return db_obj
    
    async def create_if_not_exists(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType, 
        index_elements: Optional[List[str]] = None,
        exclude_attrs: Optional[List[str]] = None,
        **kwargs,
    ) -> None:
        """
        Creates an object if it does not exist
        """
        values = self.prepare_encoded_object(obj_in, method = 'create_if_not_exists', exclude = exclude_attrs, **kwargs)
        stmt = insert(self.model).values(
            **values
        ).on_conflict_do_nothing(
            index_elements=index_elements
        )
        await db.execute(stmt)
        if self.auto_commit:
            await db.commit()
    
    async def exists(
        self,
        db: AsyncSession,
        id: Any,
        key: Optional[str] = None,
        **kwargs,
    ) -> bool:
        """
        Checks if the object exists
        """
        key = key or self.table.primary_key.columns[0].name
        stmt = select(self.model).where(getattr(self.model, key) == id)
        return await db.scalar(stmt) is not None

    """
    Upsert Operations
    """

    async def upsert(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType, 
        index_elements: Optional[List[str]] = None,
        exclude_attrs: Optional[List[str]] = None,
        no_commit: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Upserts an object
        """
        index_elements = index_elements or ['id']
        values = self.prepare_encoded_object(obj_in, method = 'upsert', exclude = exclude_attrs, **kwargs)
        if hasattr(self.model, 'updated_at'):
            values['updated_at'] = datetime.datetime.now(datetime.timezone.utc)
        # from lazyops.utils.logs import logger
        # logger.info(f'Upserting: {values}', prefix = f'|g|{self.model.__name__}|e|', colored = True)
        stmt = insert(self.model).values(
            **values
        ).on_conflict_do_update(
            index_elements=index_elements,
            set_=values
        )
        await db.execute(stmt)
        if self.auto_commit and not no_commit:
            await db.commit()
    
    async def upsert_many(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
    ):
        """
        Upserts multiple objects
        """
        for obj_in in objs_in:
            await self.upsert(db, obj_in = obj_in, no_commit = True)
        await db.commit()

    async def upsert_bulk(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        batch_size: Optional[int] = 200,
    ) -> int:
        """
        Upserts multiple objects
        """
        for i in range(0, len(objs_in), batch_size):
            await self.upsert_many(db, objs_in = objs_in[i:i+batch_size])

    
    async def upsert_many_with_idx(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        start_idx: int,
        raise_errors: Optional[bool] = None,
    ) -> Tuple[List[int], List[int]]:
        """
        Upserts multiple objects with an index
        """
        completed_idx: List[int] = []
        failed_idx: List[int] = []
        for n, obj_in in enumerate(objs_in):
            try:
                await self.upsert(db, obj_in = obj_in, no_commit = True)
                completed_idx.append(start_idx + n)
            
            except Exception as e:
                self.logger.info(f'|r|Attempting to rollback on error|e|: {e}', colored = True, prefix = f'|r|[{start_idx + n}] Error in upserting data|e|')
                await db.rollback()
                try:
                    await self.upsert(db, obj_in = obj_in, no_commit = True)
                    completed_idx.append(start_idx + n)
                except Exception as e:
                    self.logger.info(obj_in, colored = True, prefix = f'|r|[{start_idx + n}] Error in upserting data|e|')
                    self.logger.trace('Error in upserting data', e)
                    if raise_errors: raise e
                    failed_idx.append(start_idx + n)

        await db.commit()
        return (completed_idx, failed_idx)
    

    async def upsert_bulk_with_index(
        self,
        db: AsyncSession,
        *,
        objs_in: List['SourceSchemaType'],
        batch_size: Optional[int] = 200,
        raise_errors: Optional[bool] = None,
    ) -> Tuple[List[int], List[int]]:
        """
        Upserts multiple objects
        """
        completed_idx: List[int] = []
        failed_idx: List[int] = []
        for i in range(0, len(objs_in), batch_size):
            idxs = await self.upsert_many_with_idx(db, objs_in = objs_in[i:i+batch_size], start_idx = i, raise_errors = raise_errors)
            completed, failed = idxs
            if completed: completed_idx += completed
            if failed: failed_idx += failed
        return (completed_idx, failed_idx)
    
    """
    [v2] Upsert Operations
    """

    def get_update_columns(
        self,
        exclude_attrs: Optional[List[str]] = None,
        **kwargs,
    ) -> List[str]:
        """
        Returns the update columns
        """
        exclude_attrs = exclude_attrs or []
        if hasattr(self.model, 'created_at'): exclude_attrs.append('created_at')
        return [
            c.name for c in self.table.c
            if c not in list(self.table.primary_key.columns)
            and c.name not in exclude_attrs
        ]

    async def _upsert(
        self, 
        db: AsyncSession, 
        *, 
        obj_in: SourceSchemaType, 
        index_elements: Optional[List[str]] = None,
        exclude_attrs: Optional[List[str]] = None,
        no_commit: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        [V2 Logic] Upserts an object
        """
        
        index_elements = index_elements or self.table.primary_key.columns
        update_cols = self.get_update_columns(exclude_attrs = exclude_attrs, **kwargs)
        values = await self.prepare_encoded_object_one(obj_in, method = 'upsert', exclude = exclude_attrs, **kwargs)
        stmt = insert(self.model).values(values)
        on_conflict_stmt = stmt.on_conflict_do_update(
            index_elements=index_elements,
            set_ = {k: getattr(stmt.excluded, k) for k in update_cols},
        )
        await db.execute(on_conflict_stmt)
        if self.auto_commit and not no_commit:
            await db.commit()

    async def _upsert_batch(
        self, 
        db: AsyncSession, 
        *, 
        objs_in: List[SourceSchemaType], 
        index_elements: Optional[List[str]] = None,
        exclude_attrs: Optional[List[str]] = None,
        no_commit: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        [V2 Logic] Upserts an object
        """
        index_elements = index_elements or self.table.primary_key.columns
        update_cols = self.get_update_columns(exclude_attrs=exclude_attrs, **kwargs)
        values = await self.batch_prepare_encoded_objects(objs_in, method = 'upsert', exclude = exclude_attrs, **kwargs)
        stmt = insert(self.model).values(values)
        on_conflict_stmt = stmt.on_conflict_do_update(
            index_elements = index_elements,
            set_ = {k: getattr(stmt.excluded, k) for k in update_cols},
        )
        await db.execute(on_conflict_stmt)
        if self.auto_commit and not no_commit:
            await db.commit()
    

    """
    Update Operations
    """

    async def update(
        self,
        db: AsyncSession,
        *,
        db_obj: ModelType,
        obj_in: Union[SourceSchemaType, Dict[str, Any]],
    ) -> ModelType:
        """
        Updates an object
        """
        update_data = prepare_update_data(self.model, obj_in)
        for field, value in update_data.items():
            setattr(db_obj, field, value)
        db.add(db_obj)
        if self.auto_commit:
            await db.commit()
            await db.refresh(db_obj)
        return db_obj


    async def update_by_filter(
        self, 
        db: AsyncSession, 
        obj_in: Union[SourceSchemaType, Dict], 
        **filters
    ) -> Dict:
        """
        Updates an object by filter
        """
        update_data = prepare_update_data(self.model, obj_in)
        and_stmt = self._construct_and_statement(filters)
        returning_stmt = [getattr(self.model, attr) for attr in self._usable_attributes()]
        stmt = update(self.model).where(and_stmt).values(update_data).returning(*returning_stmt)
        res = (await db.execute(stmt)).first()
        self._raise_for_not_found(res, raise_for_not_found=True)
        if self.auto_commit: await db.commit()
        return dict(zip(self._usable_attributes(), res))
    
    """
    Delete Operations
    """

    async def remove(self, db: AsyncSession, *, id: Union[int, str]) -> ModelType:
        """
        Removes an object
        """
        obj = await self.get(db, id=id, raise_for_not_found=True)
        await db.delete(obj)
        await db.commit()
        return obj

    async def remove_obj(self, db: AsyncSession, *, db_obj: ModelType) -> ModelType:
        """
        Removes an object
        """
        await db.delete(db_obj)
        await db.commit()
        return db_obj

    async def delete_all_filtered(self, db: AsyncSession, **filters) -> None:
        """
        Deletes all objects with the given filters
        """
        and_stmt = self._construct_and_statement(filters)
        stmt = delete(self.model).where(and_stmt)
        await db.execute(stmt)
        await db.commit()

    """
    Misc Operations
    """

    def _add_related_obj_loading(self, stmt: Select, load_children: List[str] = None) -> Select:
        """
        Constructs a select statement with the given related objects
        """
        if load_children:
            options = iter(getattr(self.model, attr) for attr in load_children)
            stmt = stmt.options(selectinload(*options))
        return stmt

    def _raise_for_not_found(self, obj: Optional[ModelType], raise_for_not_found: bool) -> None:
        """
        Raises an error if the object is not found
        """
        if not obj and raise_for_not_found:
            raise errors.DatabaseItemNotFoundException(self.object_class_name)

    def _construct_and_statement(self, filters: dict) -> ColumnElement:
        """
        Constructs an AND statement from the given filters
        """
        _filters = ()
        for attr, value in filters.items():
            _filters += (getattr(self.model, attr) == value,)
        return and_(*_filters)
    

    def _defer_attrs(self, stmt: Union[Select, Update, Insert], defer_attrs: Optional[List[str]]):
        """
        Returns the statement with the deferred attributes
        """
        defer_attrs = defer_attrs or []
        defer_attrs += self.defer_attrs
        if defer_attrs:
            stmt = stmt.options(*[defer(getattr(self.model, attr), raiseload=True) for attr in defer_attrs])
        return stmt

    def _usable_attributes(self) -> Generator[str, None, None]:
        """
        Returns the usable attributes
        """
        for attr in inspect(self.model).c.keys():
            if attr not in self.defer_attrs:
                yield attr


    async def _cast_results_to_objects(
        self,
        db: AsyncSession,
        *,
        results: List[Tuple[str, Any]],
        cast_to_source: Optional[bool] = None,
        **kwargs,
    ) -> List['ResultObject']:
        """
        Casts the search results to objects
        """
        object_ids = [result[0] for result in results]
        objects: List['SourceSchemaType'] = await self.get_many_by_ids(db, ids = object_ids, **kwargs)
        if cast_to_source:
            objects = [obj.model_dump(mode = 'object') for obj in objects]
        return objects


"""
Exceptions
"""

import traceback
from lazyops.libs import lazyload
from typing import List, Optional

if lazyload.TYPE_CHECKING:
    import fastapi
    # from fastapi import HTTPException
else:
    fastapi = lazyload.LazyLoad("fastapi")
    # HTTPException = lazyload.LazyLoad("fastapi.HTTPException")

class AuthZeroException(fastapi.HTTPException):
    """
    Base Auth Zero Exception
    """

    base: Optional[str] = None
    concat_detail: Optional[bool] = True
    verbose: Optional[int] = None
    default_status_code: Optional[int] = 400

    log_error: Optional[bool] = False # Always log the error message
    log_devel: Optional[bool] = False # Log the error message if in development mode

    def __init__(self, detail: str = None, error: Optional[Exception] = None, status_code: Optional[int] = None, **kwargs):
        """
        Constructor
        """
        from ..configs import settings
        message = ""
        if not detail or self.concat_detail:
            message += f"{self.base}: " or ''
        if detail: message += detail
        if error and self.verbose is not None:
            if self.verbose >= 1:
                message += f"\nError: {error}"
            elif self.verbose <= 5:
                message += f"\nError: {error}\nTraceback: {traceback.format_exc()}"
        status_code = status_code or self.default_status_code
        super().__init__(status_code=status_code, detail=message, **kwargs)
        if self.log_error or (self.log_devel and settings.is_development_env):
            self.display(message, status_code)

    def display(self, message: str, status_code: int):
        """
        Displays the error
        """
        from ..utils.lazy import logger
        logger.error(f'[{self.__class__.__name__} - {status_code}]: {message}')


class ExpiredJWTException(AuthZeroException):
    """
    Expired JWT Exception
    """
    base = "Expired JWT"
    concat_detail = True
    log_devel = True
    default_status_code = 401

class InvalidJWTException(AuthZeroException):
    """
    Invalid JWT Exception
    """
    base = "Invalid JWT"
    concat_detail = True
    log_devel = True
    default_status_code = 401

class InvalidOperationException(AuthZeroException):
    """
    Invalid Operation Exception
    """
    base = "Invalid Operation"
    concat_detail = True
    log_devel = True
    default_status_code = 400

class InvalidTokenException(AuthZeroException):
    """
    Invalid Token Exception
    """
    base = "Invalid Token"
    concat_detail = True
    log_devel = True
    default_status_code = 401

class NoTokenException(AuthZeroException):
    """
    No Token Exception
    """
    base = "Not Authorized: no auth token found"
    concat_detail = True
    log_devel = True
    default_status_code = 401


class NoAPIKeyException(AuthZeroException):
    """
    No API Key Exception
    """
    base = "Not Authorized"
    concat_detail = True
    log_devel = True
    default_status_code = 401

    def __init__(self, key: str = 'x-api-key', detail: str = None, **kwargs):
        """
        Constructor
        """
        self.base += f' no `{key}` found'
        super().__init__(detail = detail, **kwargs)

class InvalidAPIKeyException(AuthZeroException):
    """
    Invalid API Key Exception
    """
    base = "Invalid API Key"
    concat_detail = True
    log_devel = True
    default_status_code = 401

class InvalidAPIKeyPrefixException(AuthZeroException):
    """
    Invalid API Key Prefix Exception
    """
    base = "Invalid API Key Prefix. Your API Key may be deprecated. Please regenerate your API Key by logging in again"
    concat_detail = True
    log_devel = False
    default_status_code = 401

class InvalidAPIKeySecretException(AuthZeroException):
    """
    Invalid API Key Secret Exception
    """
    base = "Invalid API Key. Your API Key may not belong to the correct application / environment. Please regenerate your API Key by logging in again"
    concat_detail = True
    log_devel = True
    default_status_code = 401

class InsufficientPermissionsException(AuthZeroException):
    """
    Insufficient Permissions Exception
    """
    base = "Insufficient Permissions"
    concat_detail = True
    log_devel = True
    default_status_code = 403

class InvalidAuthorizationException(AuthZeroException):
    """
    Invalid Authorization Exception
    """
    base = "Invalid Authorization"
    concat_detail = True
    log_devel = True
    default_status_code = 403

class ExpiredSesssionException(AuthZeroException):
    """
    Expired Session Exception
    """
    base = "Session Expired. Please log in again"
    concat_detail = True
    log_devel = True
    default_status_code = 403

class AuthorizationException(AuthZeroException):
    """
    Authorization Exception
    """
    base = "Authorization Failed"
    concat_detail = True
    log_devel = True
    default_status_code = 403


class InvalidRoleException(AuthZeroException):
    """
    Invalid Role Exception
    """
    base = "User does not have role"
    concat_detail = False
    log_devel = True
    default_status_code = 403

    def __init__(self, role: str):
        """
        Initializes the invalid role exception
        """
        detail = f'User does not have role `{role}`'
        super().__init__(detail = detail)

class InvalidRolesException(AuthZeroException):
    """
    Invalid Roles Exception
    """
    base = "User does not have sufficient roles"
    concat_detail = False
    log_devel = True
    default_status_code = 403

    def __init__(self, roles: List[str], require_all: bool = False):
        """
        Initializes the invalid roles exception
        """
        detail = f'User does not have sufficient roles `{roles}`' if require_all else f'User does not have any of the roles `{roles}`'
        super().__init__(detail = detail)


class InvalidAuthRoleException(AuthZeroException):
    """
    Invalid Auth Role Exception
    """
    base = "Invalid Auth Role"
    concat_detail = True
    log_devel = True
    default_status_code = 403

class NoUserException(AuthZeroException):
    """
    No User Exception
    """
    base = "Unauthorized: no user found"
    concat_detail = True
    log_devel = True
    default_status_code = 401

class InvalidAPIKeyException(AuthZeroException):
    """
    Invalid API Key Exception
    """
    base = "Invalid API Key"
    concat_detail = True
    log_devel = True
    default_status_code = 403

class ExpiredAPIKeyData(AuthZeroException):
    """
    Expired API Key Data
    """
    
    base = "Expired API Key Data"
    concat_detail = True
    log_devel = True
    default_status_code = 403


from __future__ import annotations

"""
Parser ABC
"""

import argparse
from typing import Any, Dict, List, Optional, Tuple, Union, Type, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from pydantic import BaseModel
    from pydantic.fields import FieldInfo

class ParserException(Exception):
    """
    Parser Exception
    """

    def __init__(self, message: str, help_text: str = None):
        self._error_message = message
        if help_text: message = f'{message}\n\n{help_text}'
        super().__init__(message)
        self._help_text = help_text

class ArgParser(argparse.ArgumentParser):


    def print_help(self, file=None):
        """
        The default print_help method sends output to stdout, which isn't
        useful when run on the server. We have this raise the same exception
        as ``error`` so it gets handled in the same way (hopefully with
        help text going back to the user.)
        """
        raise ParserException('', help_text = self.format_help())

    def error(self, message):
        """error(message: string)

        Prints a usage message incorporating the message to stderr and
        exits.

        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        # args = {'prog': self.prog, 'message': message}
        # raise ValueError('%(prog)s: error: %(message)s\n' % args)
        raise ParserException(message, help_text = self.format_help())

_parsed_parser_objects: Dict[str, ArgParser] = {}

_parser_types: Dict[str, Type] = {
    'str': str,
    'int': int,
    'float': float,
    'bool': bool,
    'list': list,
    'dict': dict,
    'List': List,
    # 'tuple': tuple,
    # 'None': None,
}

def parse_one(
    name: str,
    field: FieldInfo,
) -> Tuple[Set, Dict]:
    """
    Parse a single field into the args and kwargs for the add_argument method
    """
    args = set()
    kwargs = {}

    field_type = field.annotation
    for pt in _parser_types.values():
        if field_type == Optional[pt]:
            field_type = pt
            break

    if field.is_required():
        args.add(name)
    else:
        args.add(f'--{name}')
        if field.json_schema_extra and field.json_schema_extra.get('alt'):
            alt = field.json_schema_extra['alt']
            if isinstance(alt, str): alt = [alt]
            for a in alt:
                if not isinstance(a, str): a = str(a)
                if not a.startswith('-'): a = f'-{a}'
                args.add(a)
            # args.add(f'-{field.json_schema_extra["alt"]}')
        
        elif len(name) > 3:
            if '_' in name:
                parts = name.split('_')
                short_name = ''.join([p[0] for p in parts])
                args.add(f'-{short_name}')
            else: args.add(f'-{name[:2]}')


    if field_type == bool:
        kwargs['action'] = 'store_true' if field.default is False or field.default is None else 'store_false'
    elif field_type in {list[str], list[int], list[float], List[str], List[int], List[float]}:
        kwargs['action'] = 'append'
    else:
        kwargs['type'] = field_type
    if field.default is not None:
        kwargs['default'] = field.default
    elif field.default_factory is not None:
        kwargs['default'] = None

    if field.description is not None:
        kwargs['help'] = field.description

    return args, kwargs


def create_argparser_from_model(
    model: Type['BaseModel'],
    exit_on_error: Optional[bool] = False,
    parser_class: Optional[Type[ArgParser]] = None,
    **argparser_kwargs,
) -> ArgParser:
    """
    Create an Argument Parser from a Pydantic Model
    """
    global _parsed_parser_objects

    model_key = f'{model.__module__}.{model.__name__}'
    if model_key not in _parsed_parser_objects:
        if parser_class is None: parser_class = ArgParser
        doc = model.__doc__ or ''
        usage = argparser_kwargs.pop('usage', None)
        if 'Usage' in doc and not usage:
            parts = doc.split('Usage:', 1)
            doc, usage = parts[0].strip(), parts[-1].strip()
            
        parser = parser_class(description = doc, usage = usage, exit_on_error = exit_on_error, **argparser_kwargs)
        for name, field in model.model_fields.items():
            if field.json_schema_extra and field.json_schema_extra.get('hidden', False):
                continue
            args, kwargs = parse_one(name, field)
            parser.add_argument(*args, **kwargs)
        _parsed_parser_objects[model_key] = parser
    return _parsed_parser_objects[model_key]


def create_argparser_from_model_v2(
    model: Type['BaseModel'],
    exit_on_error: Optional[bool] = False,
    sub_parser: Optional[argparse.ArgumentParser] = None,
    **argparser_kwargs,
) -> ArgParser:
    """
    Create an Argument Parser from a Pydantic Model
    """
    global _parsed_parser_objects

    from pydantic import BaseModel

    model_key = f'{model.__module__}.{model.__name__}'
    if model_key not in _parsed_parser_objects:
        doc = model.__doc__ or ''
        usage = argparser_kwargs.pop('usage', None)
        if 'Usage' in doc and not usage:
            parts = doc.split('Usage:', 1)
            doc, usage = parts[0].strip(), parts[-1].strip()
            

        parser = sub_parser if sub_parser is not None else ArgParser(description = doc, usage = usage, exit_on_error = exit_on_error, **argparser_kwargs)
        # Check if there are any subparsers
        if any(
            isinstance(field.annotation, type(BaseModel))
            for field in model.model_fields.values()
        ):
            subparsers = parser.add_subparsers(dest = 'subparser')


        for name, field in model.model_fields.items():
            if field.json_schema_extra and field.json_schema_extra.get('hidden', False):
                continue
            if isinstance(field.annotation, type(BaseModel)):
                subparser = subparsers.add_parser(name, help = field.description, argument_default = field.default, exit_on_error = exit_on_error)
                create_argparser_from_model_v2(field.annotation, exit_on_error = exit_on_error, sub_parser = subparser, **argparser_kwargs)
                # subparser = create_argparser_from_model(field.annotation, exit_on_error = exit_on_error, **argparser_kwargs)
                # parser.add_subparsers(name, subparser)
                continue

            args, kwargs = parse_one(name, field)
            parser.add_argument(*args, **kwargs)
        if sub_parser is not None:
            return parser
        _parsed_parser_objects[model_key] = parser
    return _parsed_parser_objects[model_key]
    



"""
Import Handler for gspread
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import gspread
    _gspread_available = True
except ImportError:
    gspread = object
    _gspread_available = False

def resolve_gspread(
    required: bool = False,
):
    """
    Ensures that `gspread` is available
    """
    global gspread, _gspread_available
    if not _gspread_available:
        resolve_missing('gspread', required = required)
        import gspread
        _gspread_available = True
        globals()['gspread'] = gspread


def require_gspread(
    required: bool = False,
):
    """
    Wrapper for `resolve_gspread` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_gspread, 
            func = func, 
            required = required
        )
    return decorator


from __future__ import annotations

from pydantic import model_validator, PrivateAttr
from lazyops.types import BaseSettings
from lazyops.libs.proxyobj import ProxyObject
from typing import Optional


class PostHogSettings(BaseSettings):
    """
    Posthog Settings
    """

    endpoint: Optional[str] = 'https://app.posthog.com'
    enabled: Optional[bool] = None
    api_key: Optional[str] = None
    project_id: Optional[str] = None

    client_timeout: Optional[float] = 60.0

    batch_size: Optional[int] = 100 # If the total events is greater than this, it will dispatch the batch
    batch_interval: Optional[float] = 60.0 # if the wait duration is greater than this, it will dispatch the batch
    default_retries: Optional[int] = 3
    batched: Optional[bool] = True
    num_workers: Optional[int] = 1
    
    debug_enabled: Optional[bool] = None

    # The internal enabled flag without affecting the user-provided value
    _enabled: Optional[bool] = PrivateAttr(None)

    class Config:
        env_prefix = 'POSTHOG_'
        case_sensitive = False
        extra = 'allow'

    
    @model_validator(mode = 'after')
    def validate_posthog_config(self):
        """
        Validates the Posthog Configuration
        """
        self.update_enabled()
        return self
    
    def update_enabled(self):
        """
        Update the enabled status
        """
        if self.enabled is None: self._enabled = self.api_key is not None
        else: self._enabled = self.enabled
        

    @property
    def is_enabled(self) -> bool:
        """
        Returns True if the Posthog is enabled
        """
        return self._enabled

ph_settings: PostHogSettings = ProxyObject(
    obj_getter = 'lazyops.libs.posthog.utils.get_posthog_settings',
)


from __future__ import annotations

import collections.abc
from typing import Dict, Optional


def update_dict(d: Dict, u: Dict, exclude_none: Optional[bool] = False) -> Dict:
    """
    Recursively update a dictionary
    """
    for k, v in u.items():
        if exclude_none and v is None:
            continue
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        elif isinstance(v, list):
            d[k] = d.get(k, []) + v
        else:
            d[k] = v
    return d


def merge_recursive_dict(
    d: Dict,
    current_key: str,
) -> Dict:
    """
    Merge a recursive dictionary
    """
    if not isinstance(d, collections.abc.Mapping):
        return d
    mapping = {}
    for k, v in d.items():
        current_key = f'{current_key}.{k}'
        if isinstance(v, (collections.abc.Mapping, dict)):
            mapping[current_key] = merge_recursive_dict(v, current_key)
        else:
            mapping[current_key] = v
    return mapping


def flatten_dict_value(
    d: Dict,
    parent_key: str = '',
) -> str:
    """
    Flatten a dictionary
    """
    items = []
    for k, v in d.items():
        new_key = f'{parent_key}.{k}' if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict_value(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


"""
Build Script Helper

Usage:

Run a Build Step

$ lazyops-build run <step>[1-4]

Run a Pip Hotfix

$ lazyops-build hotfix <file>

Update the Build Config

$ lazyops-build config <path>

Break the Cache

$ lazyops-build breakcache

Patch the SSL Config

$ lazyops-build patchssl
"""

import os
os.environ['DEBIAN_FRONTEND'] = 'noninteractive'

import yaml
import typer
import contextlib
import subprocess
import concurrent.futures
from pathlib import Path
from pydantic import model_validator
from lazyops.types.models import BaseModel
from lazyops.libs.proxyobj import ProxyObject
from typing import Optional, List, Any, Dict, Union

class COLOR:
    """
    Color Constants
    """
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    WHITE = '\033[97m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'

# Constants / Env Vars
# BUILDS_ENABLED_SERVICES: List[str] = os.environ.get('BUILDS_ENABLED_SERVICES', 'server').split(',')
GITHUB_TOKEN: str = os.environ.get('GH_TOKEN', os.getenv('GITHUB_TOKEN', ''))

# Paths
TEMP_PATH = os.getenv('TEMP_PATH', '/tmp')
BUILD_CONFIG_PATH = Path(os.getenv('BUILD_CONFIG_PATH', f'{TEMP_PATH}/build_config.yaml'))

_HERE = Path(__file__).parent
DEFAULT_CONFIG_PATH = _HERE.joinpath('default_build_config.yaml')

APP_HOME = os.getenv('APP_HOME', '/app')
APP_PATH = Path(APP_HOME)

REQUIREMENTS_PATH = Path(os.getenv('REQUIREMENTS_PATH', '/tmp/requirements'))
PKGS_PATH = REQUIREMENTS_PATH.joinpath('pkgs')

# APP_NAME = os.getenv('APP_NAME', "App")

def echo(message: str):
    """ 
    Helper for printing a message
    """
    typer.echo(message, color = True)

# The Build Config Variables
class BuildConfig(BaseModel):
    """
    Build Config
    """
    app_name: Optional[str] = None
    builds: Optional[Dict[str, Union[str, List[str], Dict[str, Union[List[str], Dict, Any]]]]] = {}
    custom_installers: Optional[Dict[str, List[str]]] = {}
    custom_commands: Optional[Dict[str, List[str]]] = {}
    enabled_build_services: Optional[List[str]] = None
    kinds: Optional[Dict[str, Dict[str, Union[List[str], Dict[str, List[str]]]]]] = None

    @classmethod
    def load(cls, path: Optional[Path] = None) -> 'BuildConfig':
        """
        Load the build config
        """
        if path is None:
            path = BUILD_CONFIG_PATH if BUILD_CONFIG_PATH.exists() else DEFAULT_CONFIG_PATH
        return cls.parse_obj(yaml.safe_load(path.read_text()))
    
    @model_validator(mode = 'after')
    def validate_app_env(self):
        """
        Validates the app environment
        """
        if self.app_name is None: self.app_name = os.getenv('APP_NAME', "App")
        if self.enabled_build_services is None:
            if services := os.getenv('BUILDS_ENABLED_SERVICES'):
                self.enabled_build_services = services.split(',')
            else: self.enabled_build_services = list(self.builds.keys())
        if self.kinds is None:
            self.kinds = {}
            for t, conf in self.builds.items():
                if conf.get('kind'):
                    if conf['kind'] not in self.kinds:
                        self.kinds[conf['kind']] = {}
                    if t not in self.kinds[conf['kind']]:
                        self.kinds[conf['kind']][t] = []
                    if conf.get('names'):
                        self.kinds[conf['kind']][t].extend(conf['names'])
                    self.kinds[conf['kind']][t] = list(set(self.kinds[conf['kind']][t]))
        return self

    def show_env(self, step: str):
        """
        Show the environment
        """
        echo(f"Starting Step: {COLOR.GREEN}{step}{COLOR.END}\n")
        echo(f"[Enabled Builds]: {COLOR.BLUE}{self.enabled_build_services}{COLOR.END}")

    def update_config(self, path: Optional[Path] = None, **config: Any):
        """
        Update the build config
        """
        data = self.model_dump()
        if path: 
            update_data = yaml.safe_load(path.read_text())
            data.update(update_data)
        if config: data.update(config)
        BUILD_CONFIG_PATH.write_text(yaml.dump(data, default_flow_style = False))
    
    def has_service(self, service: str, fixed: Optional[bool] = True) -> bool:
        """
        Helper for checking whether the service is enabled
        """
        if not fixed:
            return any((name in service or service in name) for name in self.enabled_build_services)
        return any(
            name in self.enabled_build_services 
            for name in self.builds.get(service, {}).get('names', [])
        )
    
    @property
    def kinds(self) -> Dict[str, Dict[str, List[str]]]:
        """
        Get the kinds
        """
        return self.builds

    

cmd = typer.Typer(no_args_is_help = True)

def build_aliases(name: str, additional_names: Optional[List[str]] = None) -> List[str]:
    """
    Create the aliases for a given service name
    """
    aliases = [name]
    if '.' not in name:
        if '-' in name: aliases.append(name.replace('-', '.'))
        elif '_' in name: aliases.append(name.replace('_', '.'))
    if '-' not in name:
        if '.' in name: aliases.append(name.replace('.', '-'))
        elif '_' in name: aliases.append(name.replace('_', '-'))
    if '_' not in name:
        if '.' in name: aliases.append(name.replace('.', '_'))
        elif '-' in name: aliases.append(name.replace('-', '_'))
    if additional_names: aliases.extend(additional_names)
    return list(set(aliases))

# service_names: Dict[str, List[str]] = {}
# kinds: Dict[str, Dict[str, List[str]]] = {}

config: BuildConfig = ProxyObject(obj_getter = BuildConfig.load)

def add_to_env(
    envvar: str,
    envval: Any
):
    """
    Helper for adding to the environment variable
    """
    envval = str(envval)
    if ' ' in envval: envval = f'"{envval}"'
    os.system(f"echo 'export {envvar}={envval}' >> ~/.bashrc")
    os.environ[envvar] = envval


def parse_text_file(path: Path) -> List[str]:
    """
    Parses a text file
    """
    text_lines = path.read_text().split('\n')
    return [line.strip() for line in text_lines if ('#' not in line[:5] and line.strip())]

def get_apt_packages(kind: str, name: str) -> List[str]:
    """
    Helper for getting the apt packages for a given service/builder
    """
    # pkg_dir = PKGS_PATH.joinpath(kind)
    pkg_dir = PKGS_PATH
    for alias in config.kinds[kind].get(name, []):
        pkg_file = pkg_dir.joinpath(f'{alias}.txt')
        if pkg_file.exists():
            return parse_text_file(pkg_file)
    return []



def get_pip_requirements(kind: str, name: str) -> Optional[str]:
    """
    Helper for getting the pip requirements file for a given service/builder
    """
    pkg_dir = REQUIREMENTS_PATH.joinpath(kind)
    for alias in config.kinds[kind].get(name, []):
        req_file = pkg_dir.joinpath(f'{alias}.txt')
        if req_file.exists():
            req_file.write_text(req_file.read_text().replace('GITHUB_TOKEN', GITHUB_TOKEN))
            return req_file.as_posix()
        req_file = pkg_dir.joinpath(f'requirements.{alias}.txt')
        if req_file.exists():
            return req_file.as_posix()
    return None


"""
Step 1: Install Apt Packages
"""

APT_PKGS = []

def add_to_apt_pkgs(*pkg: str):
    """
    Helper for adding to the apt packages
    """
    global APT_PKGS
    APT_PKGS.extend(pkg)


def run_apt_install():
    """
    Helper for running apt install
    """
    if not APT_PKGS: return
    _pkgs = ' '.join(list(set(APT_PKGS)))
    echo(f"Installing Apt Packages: {COLOR.BLUE}{_pkgs}{COLOR.END}")
    os.system(f"apt-get update && apt-get -yq install --no-install-recommends {_pkgs}")

def run_cmd(cmdstr: str):
    """
    Helper for running a command
    """
    try:
        data = subprocess.check_output(cmdstr, shell=True, text=True, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
        if data[-1:] == '\n': data = data[:-1]
        echo(data)

    except subprocess.CalledProcessError as e:
        echo(f'{COLOR.RED}Failed to run command: {cmdstr}{COLOR.END}')
        echo(f'{COLOR.RED}Error: {e.stderr}{COLOR.END}')
        raise e


def get_apt_pkg_requirements():
    """
    Helper for getting the apt package requirements
    """
    for service in config.builds:
        if config.has_service(service):
            if service_pkgs := get_apt_packages(config.builds[service]['kind'], service):
            # if service_pkgs := get_apt_packages(service):
                echo(f'{COLOR.BLUE}[{config.builds[service]['kind']}]{COLOR.END} Adding {COLOR.BOLD}{service}{COLOR.END} requirements\n\n - {COLOR.BOLD}{service_pkgs}{COLOR.END}\n')
                add_to_apt_pkgs(*service_pkgs)



# Custom Requirements for Services / Builders

def run_custom_install_requirements(
    name: str,
):
    """
    Helper for running the custom install requirements
    """
    if custom_installers := config.builds.get(name, {}).get('custom_install'):
        echo(f'{COLOR.BLUE}[{name}]{COLOR.END} Running {COLOR.BOLD}Custom Installers{COLOR.END}')
        for custom in custom_installers:
            if custom_install := config.custom_installers.get(custom):
                echo(f'{COLOR.BLUE}[{name}]{COLOR.END} Running {COLOR.BOLD}{custom}{COLOR.END}')
                for cmdstr in custom_install:                    
                    os.system(cmdstr)


def run_step_one():
    """
    Step 1: Install Apt Packages
    """
    config.show_env(f'Step 1: {config.app_name} Package Installations')
    get_apt_pkg_requirements()
    run_apt_install()
    for service in config.enabled_build_services:
        run_custom_install_requirements(service)
    echo(f"{COLOR.GREEN}Step 1: {config.app_name} Package Installations Complete{COLOR.END}\n\n")


"""
END Step 1: Install Apt Packages
"""


"""
START Step 2: Install Pip Requirements
"""

def run_step_two():
    """
    Function for installing stage two requirements
    """
    config.show_env(f'Step 2: {config.app_name} Pip Installations')
    service_names = config.enabled_build_services
    for service in service_names:
        kind = config.builds[service]['kind']
        echo(f'{COLOR.BLUE}[{kind}]{COLOR.END} Installing {COLOR.BOLD}{service}{COLOR.END} requirements')
        if req_file := get_pip_requirements(kind, service):
            os.system(f"pip install -r {req_file}")
            if 'custom_commands' in config.builds[service]:
                for custom in config.builds[service]['custom_commands']:
                    if custom_cmd := config.custom_commands.get(custom):
                        echo(f'{COLOR.BLUE}[{kind}]{COLOR.END} Running {COLOR.BOLD}{custom}{COLOR.END}')
                        for cmdstr in custom_cmd:
                            os.system(cmdstr)

    echo(f"{COLOR.GREEN}Step 2: {config.app_name} Pip Installations Complete{COLOR.END}\n\n")

"""
END Step 2: Install Pip Requirements
"""

"""
START Step 3: Run Post-Installation Requirements
"""

def init_build_service(
    name: str,
):
    """
    Helper for initializing the build service
    """
    enabled = config.has_service(name)
    kind = config.builds[name]['kind']
    if enabled:
        echo(f'{COLOR.BLUE}[{kind}]{COLOR.END} Initializing {COLOR.BOLD}{name}{COLOR.END}')
        if config.builds[name].get('init'):
            init_cmd = config.builds[name]['init']
            from lazyops.utils.lazy import lazy_import
            init_func = lazy_import(init_cmd)
            init_func()
        elif name == 'server':
            with contextlib.suppress(Exception):
                from lazyops.utils.lazy import lazy_import
                init_func = lazy_import(f'{config.app_name}.app.main.init_build')
                init_func()
        echo(f'{COLOR.BLUE}[{kind}]{COLOR.END} Completed {COLOR.BOLD}{name}{COLOR.END} Initialization')
        add_to_env(f'{name.upper()}_{kind.upper()}_ENABLED', True)
    else:
        add_to_env(f'{name.upper()}_{kind.upper()}_ENABLED', False)


def init_build_deps():  # sourcery skip: extract-method
    """
    Initialize the worker dependencies

    - Do this multi-threaded later
    """
    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = [
            executor.submit(init_build_service, name) for name in config.enabled_build_services
        ]
        for future in concurrent.futures.as_completed(futures):
            _ = future.result()
    

def run_step_three():
    """
    Function for running the third step
    """
    config.show_env(f'Step 3: {config.app_name} Post-Installations')
    init_build_deps()
    echo(f"{COLOR.GREEN}Step 3: {config.app_name} Post-Installations Complete{COLOR.END}\n\n")

"""
END Step 3: Run Post-Installation Requirements
"""

"""
START Step 4: Finalize and Test
"""

def run_validate_step(
    name: str,
):
    """
    Helper for running the validation step
    """
    kind = config.builds[name]['kind']
    echo(f'{COLOR.BLUE}[{kind}]{COLOR.END} Validating {COLOR.BOLD}{name}{COLOR.END}')
    if config.builds[name].get('validate'):
        validate_cmd = config.builds[name]['validate']
        from lazyops.utils.lazy import lazy_import
        validate_func = lazy_import(validate_cmd)
        validate_func()
    elif name == 'server':
        with contextlib.suppress(Exception):
            from lazyops.utils.lazy import lazy_import
            validate_func = lazy_import(f'{config.app_name}.app.main.validate_build')
            validate_func()

    echo(f'{COLOR.BLUE}[{kind}]{COLOR.END} Completed {COLOR.BOLD}{config.app_name}{COLOR.END} Validation')


def run_step_four():
    """
    Function for running the fourth step
    """
    config.show_env(f'Step 4: {config.app_name} Installation & Validation')
    os.system(f'pip install {APP_PATH}')
    Path('/data').mkdir(exist_ok = True)
    add_to_env(f'{config.app_name.upper()}_DATA_DIR', '/data')
    
    import time
    add_to_env(f'{config.app_name.upper()}_BUILD_DATE', int(time.time()))

    for t in config.enabled_build_services:
        names = config.builds[t].get('names', [t])
        for name in names:
            if config.has_service(name):
                run_validate_step(name, fixed = name == 'server')
    echo(f"{COLOR.GREEN}Step 4: {config.app_name} Validation Complete{COLOR.END}\n\n")


@cmd.command("run")
def run_build(
    step: int = typer.Argument(1, help = "Step to Run", show_default = True, min = 1, max = 4),
):
    """
    Usage:

    Run a Scout Build Step
    $ run <step>
    """
    if step == 1:
        run_step_one()
    elif step == 2:
        run_step_two()
    elif step == 3:
        run_step_three()
    elif step == 4:
        run_step_four()
    else:
        typer.echo(f"Invalid Step: {step}")
    
@cmd.command("hotfix")
def run_pip_hotfix(
    file: Path = typer.Argument(..., help = "Filename to Hotfix", resolve_path = True),
):
    """
    Runs hotfix for pip requirements
    """
    if not file.exists():
        echo(f'{COLOR.RED}File does not exist: {file.as_posix()}{COLOR.END}')
        raise ValueError(f'File does not exist: {file.as_posix()}')
    
    echo(f'{COLOR.BLUE}Running Hotfix for {file.as_posix()}{COLOR.END}')
    reqs = parse_text_file(file)
    for req in reqs:
        echo(f'{COLOR.BLUE}Installing: {req}{COLOR.END}')
        if 'GITHUB_TOKEN' in req:
            req = req.replace('GITHUB_TOKEN', GITHUB_TOKEN)
        os.system(f'pip install --upgrade --no-deps --force-reinstall "{req}"')
    

@cmd.command("config")
def run_update_build_config(
    path: Path = typer.Argument(BUILD_CONFIG_PATH, help = "Path to Build Config", resolve_path = True),
):
    """
    Runs the build config
    """
    if not path.exists():
        echo(f'{COLOR.RED}File does not exist: {path.as_posix()}{COLOR.END}')
        raise ValueError(f'File does not exist: {path.as_posix()}')
    echo(f'{COLOR.BLUE}Updating Build Config: {path.as_posix()}{COLOR.END}')
    config.update_config(path)


@cmd.command("breakcache")
def run_break_cache():
    """
    Runs the break cache command
    """
    os.system('echo "Breaking Cache: $(date)" > /tmp/.breakcache')

@cmd.command("patchssl")
def run_patch_ssl():
    """
    Patch the OpenSSL Config for Python
    """
    openssl_path = APP_PATH.joinpath('openssl.cnf')
    echo(f'{COLOR.BLUE}Patching OpenSSL Config: {openssl_path.as_posix()}{COLOR.END}')
    openssl_path.write_text("""
openssl_conf = openssl_init

[openssl_init]
ssl_conf = ssl_sect

[ssl_sect]
system_default = system_default_sect

[system_default_sect]
Options = UnsafeLegacyRenegotiation
""".strip())
    add_to_env('OPENSSL_CONF', openssl_path.as_posix())

def main():
    """
    Main Function
    """
    cmd()

if __name__ == "__main__":
    main()



from .oauth import AuthZeroOAuthClient
from .api_client import AuthZeroAPIClient
from .dependencies import (
    CurrentUser,
    OptionalUser, 
    ValidUser, 
    get_current_user, 
    require_auth_role,
    require_roles,
    require_api_key,
    require_api_key_or_user_role,

    auth_role_dependency,
    api_key_or_user_role_dependency,
)


import contextlib
import asyncio

def is_in_async_loop() -> bool:
    """
    Returns True if the function is called in an async loop
    """
    with contextlib.suppress(Exception):
        return asyncio.get_event_loop().is_running()
    return False




from __future__ import annotations

"""
Env Var Helpers
"""

import os
import re
import json
from pathlib import Path
from lazyops.utils.helpers import build_dict_from_str
from typing import Any, Dict, List, Optional, Tuple, Union, Type, TypeVar, TYPE_CHECKING


EnvT = TypeVar('EnvT')

def parse_from_envvar(
    var: str,
    default: Optional[Any] = None,
    _type: Optional[Type[EnvT]] = str,
) -> EnvT:
    """
    Parse a value from an environment variable
    """
    val = os.getenv(var)
    if val is None: return default

    # Try to evaluate the value
    if _type is None: 
        # print('No type: ', val)
        return val
        # if val.startswith(('{', '[')):
        #     return json.loads(val)
        # return build_dict_from_str(val) if val.startswith('{') else val.split(',')
    if _type in {list, dict}:
        if val.startswith(('{', '[')):
            return json.loads(val)
        return build_dict_from_str(val) if _type is dict else val.split(',')

    if _type is Path:
        return Path(val)
    
    if _type is bool:
        return val.lower() in {'true', '1', 't', 'y', 'yes'}
    
    return _type(val)


def parse_envvars_from_text(
    text: str,
    values: Optional[Dict[str, Any]] = None,
    envvar_prefix: Optional[str] = 'env/',
) -> Tuple[str, Dict[str, Any]]:
    """
    Parse values from a text block

    Returns:
        Tuple[str, Dict[str, Any]]: The text with the envvars replaced and the parsed values
    """

    # Create a pattern that would match env/ENVVAR_NAME and capture the ENVVAR_NAME
    _prefix = envvar_prefix.replace('/', '\/')
    pattern = re.compile(rf'({_prefix}\w+)')

    # Find all
    values = values or {}
    matches = pattern.findall(text)
    for match in matches:
        # var = match[0]
        envvar = match.replace(envvar_prefix, '')
        # print(match, envvar)
        _default = values.get(envvar)
        if isinstance(_default, type):
            _type = _default
            _default = None
        else:
            _type = type(_default) if _default is not None else None
        val = parse_from_envvar(envvar, default=_default, _type=_type)
        # if val is not None:
        values[envvar] = val
        # print(match, val, values)
        text = text.replace(match, val if val is not None else '')

    return text, values




from __future__ import annotations

"""
API Key Data Flows
"""
import time
from ..types.properties import StatefulProperty
from ..types.auth import APIKeyData
from .admin import AZManagementClient, logger
from ..utils.lazy import get_az_flow

from typing import List, Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..types.claims import UserJWTClaims
    from ..types.user_data import AZUserData
    from ..flows.user_data import UserDataFlow

class APIKeyDataFlow(StatefulProperty[APIKeyData]):
    """
    API Key Data Flow
    """
    name: Optional[str] = "api_keys"
    schema_type = APIKeyData

    def __init__(
        self,
        user_id: str,
        **kwargs,
    ):
        cache_key = self.create_hash_key(user_id) if self.settings.session_user_id_hashed else user_id
        super().__init__(
            cache_key = cache_key,
            **kwargs,
        )
        self.user_id = user_id
        self.mtg = AZManagementClient
        self._user_data_flow: 'UserDataFlow' = None

    @property
    def user_data_flow(self) -> 'UserDataFlow':
        """
        Returns the User Data Flow
        """
        if self._user_data_flow is None:
            self._user_data_flow = get_az_flow('user_data', user_id = self.user_id)
        return self._user_data_flow

    def get_ttl(self, data: APIKeyData) -> Optional[int]:
        """
        Returns the TTL
        """
        return None
        # return data.user_data.expiration_ts - int(time.time())
        # ttl = data.user_data.expiration_ts - int(time.time())
        # self.settings.autologger.info(f'Checking if API Key Data is Expired: {self.user_id}: {ttl}')
        # return ttl
        # return 30
        # return None
    
    def is_expired(self, data: APIKeyData) -> bool:
        """
        Returns True if the Data is Expired
        """
        # return data.user_data.is_expired
        return False
    
    async def afetch(self) -> APIKeyData:
        """
        Fetches the API Key Data
        """
        data: Optional['APIKeyData'] = await self.pdict.aget(self.cache_key)
        if data is not None:
            if not self.is_valid_type(data):
                await self.adelete()
                data = None
            elif data.user_data.is_expired:
                logger.info(f'Refreshing User Data: {self.user_id}')
                data.user_data = await self.user_data_flow.afetch()
                await self.asave(data)
        # if data is not None and (not self.is_valid_type(data) or self.is_expired(data)):
        #     await self.adelete()
        #     # logger.warning(f'Expired API Key Data for {self.user_id}')
        #     # logger.warning(f'Invalid API Key Data for {self.user_id}: {(type(data))}')
        #     data = None
        return data
    

    async def aset(
        self,
        user_data: 'AZUserData',
        claims: 'UserJWTClaims',
    ) -> 'APIKeyData':
        """
        Saves the Data / Claims for the User ID
        """
        data = self.schema_type(
            user_data = user_data,
            claims = claims.to_api_key_claims() if hasattr(claims, 'to_api_key_claims') else claims,
        )
        await self.pdict.aset(self.cache_key, data) # , ex = self.get_ttl(data))
        # self.settings.logger.info(f'Saved API Key Data: {self.cache_key}', colored = True, prefix = self.user_id)
        return data



    if TYPE_CHECKING:
        def fetch(self) -> APIKeyData:
            """
            Fetches the API Key Data
            """
            ...

        async def afetch(self) -> APIKeyData:
            """
            Fetches the API Key Data
            """
            ...
        



"""
Import Handler for `aiosqlite`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import aiosqlite
    _aiosqlite_available = True
except ImportError:
    aiosqlite = object
    _aiosqlite_available = False

def resolve_aiosqlite(
    required: bool = False,
):
    """
    Ensures that `aiosqlite` is available
    """
    global aiosqlite, _aiosqlite_available
    if not _aiosqlite_available:
        resolve_missing('aiosqlite', required = required)
        import aiosqlite
        _aiosqlite_available = True
        globals()['aiosqlite'] = aiosqlite


def require_aiosqlite(
    required: bool = False,
):
    """
    Wrapper for `resolve_aiosqlite` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_aiosqlite, 
            func = func, 
            required = required
        )
    return decorator


"""
Support Lazy Loading of Modules

- Credit: https://github.com/kevdog824


from lazyops.libs import lazyload

if lazyload.TYPE_CHECKING:
    import matplotlib
    import os
    import pandas as pd
else:
    matplotlib = lazyload.LazyLoad("matplotlib")
    os = lazyload.LazyLoad("os")
    pd = lazyload.LazyLoad("pandas")


def sometimes_called():
    # if this function is not called pandas library is never loaded
    pd.show_versions()


print(os)                  # prints <Uninitialized module 'os'>
os.cpu_count()             # attribute access on LazyLoad object implicitly loads module
print(os)                  # prints <module 'os' (frozen)>


print(matplotlib)          # prints <Uninitialized module 'matplotlib'>
lazyload.load(matplotlib)  # explicitly load the module backing this LazyLoad object
print(matplotlib)          # prints <module 'matplotlib' from '...'>


sometimes_called()
"""

from .main import LazyLoad, lazy_load, load
from .wrappers import lazy_function_wrapper
from typing import TYPE_CHECKING



from typing import TYPE_CHECKING, Union, Type

if TYPE_CHECKING:
    from .auth import AuthObject, AuthZeroTokenAuth, APIKeyData
    from .claims import UserJWTClaims, APIKeyJWTClaims
    from .current_user import CurrentUser
    from .user_data import AZUserData
    from .user_roles import UserRole
    from .user_session import UserSession
    
    AZResource = Union[
        APIKeyData,
        APIKeyJWTClaims,
        AuthObject,
        AuthZeroTokenAuth,
        AZUserData,
        CurrentUser,
        UserJWTClaims,
        UserSession,
    ]

    AZResourceSchema = Type[
        Union[
            APIKeyData,
            APIKeyJWTClaims,
            AuthObject,
            AuthZeroTokenAuth,
            AZUserData,
            CurrentUser,
            UserJWTClaims,
            UserSession,
        ]
    ]
